import ast
from typing import Tuple, List

from tiivad.file import File
from tiivad.t import format_error

from enum import Enum


class ValidationType(Enum):
    ALL_OF_THESE = "ALL_OF_THESE"
    ANY_OF_THESE = "ANY_OF_THESE"
    ANY = "ANY"
    NONE_OF_THESE = "NONE_OF_THESE"
    MISSING_AT_LEAST_ONE_OF_THESE = "MISSING_AT_LEAST_ONE_OF_THESE"
    NONE = "NONE"


class Program:
    def __init__(self, program_name: str):
        self.file: File = File(program_name)
        self.syntax_tree: ast.Module = None
        self.contains_loop: bool = None
        self.calls_print: bool = None
        self.contains_try_except: bool = None

    def get_syntax_tree(self) -> ast.Module:
        """
        Get AST syntax tree of the program at hand.
        :return: Syntax tree of the program at hand.
        """
        if self.syntax_tree is None:
            try:
                self.syntax_tree = ast.parse(self.file.get_file_text())
            except Exception as e:
                self.syntax_tree = ""
        return self.syntax_tree

    def program_imports_module(self, module_name: str, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if program imports module.
        :param module_name: Module name that is being checked.
        :param failed_message: The message that will be displayed in case of failing the test.
        :return: True if the program imports the specified module, False if doesn't import.
        """
        if not self.get_syntax_tree():
            return False, failed_message
        for node in ast.walk(self.get_syntax_tree()):
            if isinstance(node, ast.Import):
                if 'names' in node._fields:
                    for mod_name in node.names:
                        if 'name' in mod_name._fields:
                            if mod_name.name == module_name:
                                return True, failed_message
                            split_names = mod_name.name.split(".")
                            for split_name in split_names:
                                if split_name == module_name:
                                    return True, failed_message
            if isinstance(node, ast.ImportFrom):
                if 'module' in node._fields:
                    for name in node.names:
                        if name.name == module_name:
                            return True, failed_message
                    if node.module == module_name:
                        return True, failed_message
                    split_names = node.module.split(".")
                    for split_name in split_names:
                        if split_name == module_name:
                            return True, failed_message

        return False, format_error(failed_message, {"module_name": module_name})

    def program_imports_module2(self, validation_type: str, modules: List[str], nothing_else: bool,
                                failed_message: str = "") -> Tuple[bool, str]:
        if not self.get_syntax_tree():
            return False, failed_message

        if validation_type == ValidationType.ALL_OF_THESE.name:
            for module in modules:
                if not self.program_imports_module(module, failed_message)[0]:
                    return False, failed_message

        elif validation_type == ValidationType.ANY_OF_THESE.name:
            for module in modules:
                if self.program_imports_module(module, failed_message)[0]:
                    return True, ""
            return False, failed_message

        elif validation_type == ValidationType.ANY.name:
            if not self.program_imports_any_module(failed_message)[0]:
                return False, failed_message

        elif validation_type == ValidationType.NONE_OF_THESE.name:
            for module in modules:
                if self.program_imports_module(module, failed_message)[0]:
                    return False, failed_message

        elif validation_type == ValidationType.MISSING_AT_LEAST_ONE_OF_THESE.name:
            for module in modules:
                if not self.program_imports_module(module, failed_message)[0]:
                    return True, failed_message
            return False, failed_message

        elif validation_type == ValidationType.NONE.name:
            if self.program_imports_any_module(failed_message)[0]:
                return False, failed_message

        else:
            raise NotImplementedError(
                f"The following type is not implemented for the program_imports_module_test: {validation_type}")

        return True, ""

    def program_imports_module_from_set(self, module_names: List[str], failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if program imports any given module from the set.
        :param module_names: Module names that are being checked.
        :param failed_message: The message that will be displayed in case of failing the test.
        :return: True if the program imports any module from the set, False if it doesn't import.
        """
        if not self.get_syntax_tree():
            return False, failed_message
        for module_name in module_names:
            if self.program_imports_module(module_name)[0]:
                return True, failed_message
        return False, format_error(failed_message, {"module_name": module_names})

    def program_imports_any_module(self, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if program imports any module.
        :param failed_message: The message that will be displayed in case of failing the test.
        :return: True if imports any, False if not.
        """
        if not self.get_syntax_tree():
            return False, failed_message
        for node in ast.walk(self.get_syntax_tree()):
            if isinstance(node, ast.Import) or isinstance(node, ast.ImportFrom):
                return True, failed_message
        return False, failed_message

    def program_defines_function(self, function_name: str, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if program defines a function.
        :param failed_message: The message that will be displayed in case of failing the test.
        :param function_name: Function name that is being checked.
        :return: True if the program defines the function, False if doesn't use.
        """
        if not self.get_syntax_tree():
            return False, failed_message
        defines_function = function_name in [c.name for c in ast.walk(self.get_syntax_tree())
                                             if isinstance(c, ast.FunctionDef) and not isinstance(c, ast.Attribute)]
        if not defines_function:
            failed_message = format_error(failed_message, {"function_name": function_name})

        return defines_function, failed_message

    def program_defines_any_function(self, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if program defines any function in it.
        :param failed_message: The message that will be displayed in case of failing the test.
        :return: True if defines any other function, False if doesn't.
        """
        if not self.get_syntax_tree():
            return False, failed_message
        defines_function = len([c.name for c in ast.walk(self.get_syntax_tree())
                                if isinstance(c, ast.FunctionDef)
                                and not isinstance(c, ast.Attribute)]) > 0

        return defines_function, failed_message

    def program_contains_loop(self, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if program contains a while/for loop (including in functions).
        :return: True if contains a loop, False if not.
        """
        if not self.get_syntax_tree():
            return False, failed_message
        if self.contains_loop is None:
            self.contains_loop = False
            for node in ast.walk(self.get_syntax_tree()):
                if isinstance(node, (ast.For, ast.While, ast.comprehension)):
                    self.contains_loop = True
                    break
        return self.contains_loop, failed_message

    def program_contains_try_except(self, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if program contains a try/except block (including in functions).
        :return: True if contains a try/except block, False if not.
        """
        if not self.get_syntax_tree():
            return False, failed_message
        if self.contains_try_except is None:
            self.contains_try_except = False
            for node in ast.walk(self.get_syntax_tree()):
                if isinstance(node, (ast.Try, ast.ExceptHandler)):
                    self.contains_try_except = True
                    break
        return self.contains_try_except, failed_message

    def program_calls_function(self, function_name: str, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if program calls a given function.
        :param function_name: str, function name to be checked.
        :param failed_message: The message that will be displayed in case of failing the test.
        :return: True if program calls the given function, False if not.
        """
        result, _ = self.program_calls_function_from_set([function_name])
        if not result:
            return result, format_error(failed_message, {"function_name": function_name})
        return result, failed_message

    def program_calls_any_function(self, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if program calls any function in it.
        :param failed_message: The message that will be displayed in case of failing the test.
        :return: True if defines any other function, False if doesn't.
        """
        if not self.get_syntax_tree():
            return False, failed_message

        for c in ast.walk(self.get_syntax_tree()):
            if isinstance(c, ast.Call):
                return True, ""

        return False, failed_message

    def program_calls_function2(self, validation_type: str, modules: List[str], nothing_else: bool,
                                failed_message: str = "") -> Tuple[bool, str]:
        if not self.get_syntax_tree():
            return False, failed_message

        if validation_type == ValidationType.ALL_OF_THESE.name:
            for module in modules:
                if not self.program_calls_function(module, failed_message)[0]:
                    return False, failed_message

        elif validation_type == ValidationType.ANY_OF_THESE.name:
            for module in modules:
                if self.program_calls_function(module, failed_message)[0]:
                    return True, ""
            return False, failed_message

        elif validation_type == ValidationType.ANY.name:
            if not self.program_calls_any_function(failed_message)[0]:
                return False, failed_message

        elif validation_type == ValidationType.NONE_OF_THESE.name:
            for module in modules:
                if self.program_calls_function(module, failed_message):
                    return False, failed_message

        elif validation_type == ValidationType.MISSING_AT_LEAST_ONE_OF_THESE.name:
            for module in modules:
                if not self.program_calls_function(module, failed_message)[0]:
                    return True, failed_message
            return False, failed_message

        elif validation_type == ValidationType.NONE.name:
            if self.program_calls_any_function(failed_message)[0]:
                return False, failed_message

        else:
            raise NotImplementedError(
                f"The following type is not implemented for the program_calls_function_test: {validation_type}")

        return True, ""

    def program_defines_function2(self, validation_type: str, modules: List[str], nothing_else: bool,
                                  failed_message: str = "") -> Tuple[bool, str]:
        if not self.get_syntax_tree():
            return False, failed_message

        if validation_type == ValidationType.ALL_OF_THESE.name:
            for module in modules:
                if not self.program_defines_function(module, failed_message)[0]:
                    return False, failed_message

        elif validation_type == ValidationType.ANY_OF_THESE.name:
            for module in modules:
                if self.program_defines_function(module, failed_message)[0]:
                    return True, ""
            return False, failed_message

        elif validation_type == ValidationType.ANY.name:
            if not self.program_defines_any_function(failed_message)[0]:
                return False, failed_message

        elif validation_type == ValidationType.NONE_OF_THESE.name:
            for module in modules:
                if self.program_defines_function(module, failed_message):
                    return False, failed_message

        elif validation_type == ValidationType.MISSING_AT_LEAST_ONE_OF_THESE.name:
            for module in modules:
                if not self.program_defines_function(module, failed_message)[0]:
                    return True, failed_message
            return False, failed_message

        elif validation_type == ValidationType.NONE.name:
            if self.program_defines_any_function(failed_message)[0]:
                return False, failed_message

        else:
            raise NotImplementedError(
                f"The following type is not implemented for the program_defines_function_test: {validation_type}")

        return True, ""

    def program_calls_function_from_set(self, functions_list: list[str], failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if program calls a function from a given list.
        :param functions_list: List of function names.
        :param failed_message: The message that will be displayed in case of failing the test.
        :return: True if any of the functions in list are called, False if not.
        """
        if not self.get_syntax_tree():
            return False, failed_message
        for node in ast.walk(self.get_syntax_tree()):
            if isinstance(node, ast.Call):
                if 'func' in node._fields:
                    if 'id' in node.func._fields:
                        for i in range(len(functions_list)):
                            if node.func.id == functions_list[i]:
                                return True, failed_message
                    if 'attr' in node.func._fields:
                        for i in range(len(functions_list)):
                            if node.func.attr == functions_list[i]:
                                return True, failed_message
        return False, format_error(failed_message, {"functions_list": functions_list})

    def program_calls_print(self, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if program calls print.
        :param failed_message: The message that will be displayed in case of failing the test.
        :return: True if program calls print, False if not.
        """
        if not self.get_syntax_tree():
            return False, failed_message
        if self.calls_print is None:
            self.calls_print, _ = self.program_calls_function("print")
        return self.calls_print, failed_message

    def program_contains_keyword(self, keyword: str, failed_message: str = "") -> Tuple[bool, str]:
        """

        :param keyword:
        :param failed_message:
        :return:
        """
        if keyword in self.file.get_file_text():
            return True, ""
        else:
            return False, failed_message

    def program_contains_keyword2(self, validation_type: str, modules: List[str], nothing_else: bool,
                                  failed_message: str = "") -> Tuple[bool, str]:
        if validation_type == ValidationType.ALL_OF_THESE.name:
            for module in modules:
                if not self.program_contains_keyword(module, failed_message)[0]:
                    return False, failed_message

        elif validation_type == ValidationType.ANY_OF_THESE.name:
            for module in modules:
                if self.program_contains_keyword(module, failed_message)[0]:
                    return True, ""
            return False, failed_message

        elif validation_type == ValidationType.NONE_OF_THESE.name:
            for module in modules:
                if self.program_defines_function(module, failed_message):
                    return False, failed_message

        elif validation_type == ValidationType.MISSING_AT_LEAST_ONE_OF_THESE.name:
            for module in modules:
                if not self.program_defines_function(module, failed_message)[0]:
                    return True, failed_message
            return False, failed_message
        else:
            raise NotImplementedError(
                f"The following type is not implemented for the program_contains_keyword test: {validation_type}")

        return True, ""

    @staticmethod
    def program_output_contains_keyword(output: str, keyword: str, failed_message: str = "") -> Tuple[bool, str]:
        """

        :param output:
        :param keyword:
        :param failed_message:
        :return:
        """
        if keyword in output:
            return True, ""
        else:
            return False, failed_message

    def program_output_contains_keyword2(self, validation_type: str, expected_output: List[str], nothing_else: bool,
                                         actual_output: str, failed_message: str = "") -> Tuple[bool, str]:
        if validation_type == ValidationType.ALL_OF_THESE.name:
            for module in expected_output:
                if not self.program_output_contains_keyword(actual_output, module, failed_message)[0]:
                    return False, failed_message

        elif validation_type == ValidationType.ANY_OF_THESE.name:
            for module in expected_output:
                if self.program_output_contains_keyword(actual_output, module, failed_message)[0]:
                    return True, ""
            return False, failed_message

        elif validation_type == ValidationType.NONE_OF_THESE.name:
            for module in expected_output:
                if self.program_output_contains_keyword(actual_output, module, failed_message):
                    return False, failed_message

        elif validation_type == ValidationType.MISSING_AT_LEAST_ONE_OF_THESE.name:
            for module in expected_output:
                if not self.program_output_contains_keyword(actual_output, module, failed_message)[0]:
                    return True, failed_message
            return False, failed_message

        else:
            raise NotImplementedError(
                f"The following type is not implemented for the program_output_contains_keyword2: {validation_type}")

        return True, ""


if __name__ == '__main__':
    prog = Program("../test/samples/prog3-calc.py")
    print(prog.program_imports_module("math"))
    """
    print(prog.get_syntax_tree())
    print(prog.program_defines_function("test"))
    print(prog.program_defines_any_function())
    print(prog.program_imports_module("math"))
    print(prog.program_imports_any_module())
    print(prog.program_calls_function("test"))
    print(prog.program_calls_function("ceil"))
    print(prog.program_calls_function_from_set(["t", "test"]))
    print(prog.program_calls_function_from_set(["t", "t2"]))
    print(prog.program_contains_loop())
    print(prog.program_calls_print())
    """



import inspect
import os.path
import sys
import tempfile
from enum import Enum
from importlib.util import module_from_spec, spec_from_loader
from typing import Any, Callable

from tiivad.edutest import create_file
from tiivad.file import File
from tiivad.function import Function
from tiivad.function_execution import FunctionExecution
from tiivad.help_structures import Check
from tiivad.program import Program
from tiivad.program_execution import ProgramExecution


# TODO: Refactor so that the call would be unified for all the similar ones.
# TODO: Function contains keyword vajab töötamiseks funktsiooni objekti, mida me staatiliselt kätte ei saa?


class ShowSourceLoader:
    def __init__(self, modname: str, source: str) -> None:
        self.modname = modname
        self.source = source

    def get_source(self, modname: str) -> str:
        if modname != self.modname:
            raise ImportError(modname)
        return self.source


def make_function(s: str) -> Callable[..., Any]:
    filename = tempfile.mktemp(suffix='.py')
    modname = os.path.splitext(os.path.basename(filename))[0]
    assert modname not in sys.modules
    # our loader is a dummy one which just spits out our source
    loader = ShowSourceLoader(modname, s)
    spec = spec_from_loader(modname, loader, origin=filename)
    module = module_from_spec(spec)
    # the code must be compiled so the function's code object has a filename
    code = compile(s, mode='exec', filename=filename)
    import contextlib
    with contextlib.redirect_stdout(None):
        exec(code, module.__dict__)
    # inspect.getmodule(...) requires it to be in sys.modules
    sys.modules[modname] = module
    return module


def string_contains_string(str, str2):
    return str in str2


class TestResult(Enum):
    PASS = "PASS"
    FAIL = "FAIL"


def program_imports_module_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name
    prog = Program(program_name=kwargs["file_name"])

    for check in kwargs["standard_output_checks"]:
        test_result, error_msg = prog.program_imports_module2(check.string_check_type, check.expected_output,
                                                              check.nothing_else, check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] = kwargs["standard_output_checks"]
    return kwargs


def program_calls_function_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name
    prog = Program(program_name=kwargs["file_name"])

    for check in kwargs["standard_output_checks"]:
        test_result, error_msg = prog.program_calls_function2(check.string_check_type, check.expected_output,
                                                              check.nothing_else, check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
            break
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] = kwargs["standard_output_checks"]
    return kwargs


def program_contains_loop_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name

    prog = Program(program_name=kwargs["file_name"])
    test_result, error_msg = prog.program_contains_loop(kwargs["failed_message"])
    must_contain = kwargs["contains_check"][0]
    cannot_contain = kwargs["contains_check"][1]
    check = Check()
    check.before_message = kwargs["before_message"]
    check.passed_message = kwargs["passed_message"]
    check.failed_message = kwargs["failed_message"]
    if (must_contain and not test_result) or (cannot_contain and test_result):
        kwargs["test_status"] = TestResult.FAIL.name
        check.error_message = error_msg
        check.check_status = TestResult.FAIL.name
    else:
        check.check_status = TestResult.PASS.name

    kwargs["checks"] = [check]
    return kwargs


def program_contains_try_except_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name

    prog = Program(program_name=kwargs["file_name"])
    test_result, error_msg = prog.program_contains_try_except(kwargs["failed_message"])
    must_contain = kwargs["contains_check"][0]
    cannot_contain = kwargs["contains_check"][1]
    check = Check()
    check.before_message = kwargs["before_message"]
    check.passed_message = kwargs["passed_message"]
    check.failed_message = kwargs["failed_message"]
    if (must_contain and not test_result) or (cannot_contain and test_result):
        kwargs["test_status"] = TestResult.FAIL.name
        check.error_message = error_msg
        check.check_status = TestResult.FAIL.name
    else:
        check.check_status = TestResult.PASS.name

    kwargs["checks"] = [check]
    return kwargs


def program_calls_print_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name

    prog = Program(program_name=kwargs["file_name"])
    test_result, error_msg = prog.program_calls_print(kwargs["failed_message"])
    must_contain = kwargs["contains_check"][0]
    cannot_contain = kwargs["contains_check"][1]
    check = Check()
    check.before_message = kwargs["before_message"]
    check.passed_message = kwargs["passed_message"]
    check.failed_message = kwargs["failed_message"]
    if (must_contain and not test_result) or (cannot_contain and test_result):
        kwargs["test_status"] = TestResult.FAIL.name
        check.error_message = error_msg
        check.check_status = TestResult.FAIL.name
    else:
        check.check_status = TestResult.PASS.name

    kwargs["checks"] = [check]
    return kwargs


def program_defines_function_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name
    prog = Program(program_name=kwargs["file_name"])

    for check in kwargs["standard_output_checks"]:
        test_result, error_msg = prog.program_defines_function2(check.string_check_type, check.expected_output,
                                                                check.nothing_else, check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
            break
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] = kwargs["standard_output_checks"]
    return kwargs


def program_contains_keyword_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name
    prog = Program(program_name=kwargs["file_name"])

    for check in kwargs["standard_output_checks"]:
        test_result, error_msg = prog.program_contains_keyword2(check.string_check_type, check.expected_output,
                                                                check.nothing_else, check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
            break
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] = kwargs["standard_output_checks"]
    return kwargs


def program_execution_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name
    kwargs["checks"] = []

    for file in kwargs["input_files"]:
        create_file(file[0], file[1])

    prog = Program(program_name=kwargs["file_name"])
    user_inputs = kwargs["standard_input_data"]
    pe = ProgramExecution(prog, user_inputs)

    try:
        pe.program_execution_and_input_count()
    except:
        kwargs["exception_message"] = pe.program_error_message
        kwargs["test_status"] = TestResult.FAIL.name
        return kwargs

    kwargs["actual_output"] = pe.io_cap.get_stdout()

    # Standard output checks
    for check in kwargs["standard_output_checks"]:
        test_result, error_msg = prog.program_output_contains_keyword2(check.string_check_type, check.expected_output,
                                                                       check.nothing_else, kwargs["actual_output"],
                                                                       check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
            kwargs["checks"] += kwargs["standard_output_checks"]
            return kwargs
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] += kwargs["standard_output_checks"]

    # Output file output checks
    for check in kwargs["output_file_checks"]:
        output_file_name = check.file_name
        file_output = File(output_file_name).get_file_text()
        test_result, error_msg = prog.program_output_contains_keyword2(check.string_check_type, check.expected_output,
                                                                       check.nothing_else, file_output,
                                                                       check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
            kwargs["checks"] += kwargs["output_file_checks"]
            return kwargs
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] += (kwargs["output_file_checks"])

    threw_error = pe.program_threw_error
    must_contain = kwargs["exception_check"][0]
    cannot_contain = kwargs["exception_check"][1]

    check = Check()
    check.before_message = kwargs["before_message"]
    check.passed_message = kwargs["passed_message"]
    check.failed_message = kwargs["failed_message"]
    if (must_contain and not threw_error) or (cannot_contain and threw_error):
        kwargs["test_status"] = TestResult.FAIL.name
        check.error_message = pe.program_error_message
        check.check_status = TestResult.FAIL.name
    else:
        check.check_status = TestResult.PASS.name

    kwargs["checks"] += [check]
    return kwargs


def function_contains_loop_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name

    prog = Program(program_name=kwargs["file_name"])
    f_name = kwargs["function_name"]
    f = Function(f_name, prog.get_syntax_tree())

    test_result, error_msg = f.function_contains_loop(kwargs["failed_message"])
    must_contain = kwargs["contains_check"][0]
    cannot_contain = kwargs["contains_check"][1]

    check = Check()
    check.before_message = kwargs["before_message"]
    check.passed_message = kwargs["passed_message"]
    check.failed_message = kwargs["failed_message"]

    if must_contain and not test_result or cannot_contain and test_result:
        kwargs["test_status"] = TestResult.FAIL.name
        check.error_message = error_msg
        check.check_status = TestResult.FAIL.name
    else:
        check.check_status = TestResult.PASS.name

    kwargs["checks"] = [check]
    return kwargs


def function_contains_keyword_test(**kwargs):
    kwargs["test_status"] = TestResult.FAIL.name

    # TODO: FIXME:
    return

    prog = Program(program_name=kwargs["file_name"])
    f_name = kwargs["function_name"]
    f = Function(f_name, prog.get_syntax_tree())

    func = make_function(prog.file.get_file_text())
    fu = getattr(func, f_name)
    function_string = inspect.getsource(fu)

    for check in kwargs["standard_output_checks"]:
        test_result, error_msg = f.function_contains_keyword2(check.string_check_type, check.expected_output,
                                                              check.nothing_else, function_string, check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
            break
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] = kwargs["standard_output_checks"]
    return kwargs


def function_contains_return_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name

    prog = Program(program_name=kwargs["file_name"])
    f_name = kwargs["function_name"]
    f = Function(f_name, prog.get_syntax_tree())

    test_result, error_msg = f.function_contains_return(kwargs["failed_message"])
    must_contain = kwargs["contains_check"][0]
    cannot_contain = kwargs["contains_check"][1]

    check = Check()
    check.before_message = kwargs["before_message"]
    check.passed_message = kwargs["passed_message"]
    check.failed_message = kwargs["failed_message"]

    if must_contain and not test_result or cannot_contain and test_result:
        kwargs["test_status"] = TestResult.FAIL.name
        check.error_message = error_msg
        check.check_status = TestResult.FAIL.name
    else:
        check.check_status = TestResult.PASS.name

    kwargs["checks"] = [check]
    return kwargs


def function_calls_function_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name
    prog = Program(program_name=kwargs["file_name"])
    f_name = kwargs["function_name"]
    f = Function(f_name, prog.get_syntax_tree())

    for check in kwargs["standard_output_checks"]:
        test_result, error_msg = f.function_calls_function2(check.string_check_type, check.expected_output,
                                                            check.nothing_else, check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
            break
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] = kwargs["standard_output_checks"]
    return kwargs


def function_is_recursive_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name

    prog = Program(program_name=kwargs["file_name"])
    f_name = kwargs["function_name"]
    f = Function(f_name, prog.get_syntax_tree())

    test_result, error_msg = f.function_is_recursive(kwargs["failed_message"])
    must_contain = kwargs["contains_check"][0]
    cannot_contain = kwargs["contains_check"][1]

    check = Check()
    check.before_message = kwargs["before_message"]
    check.passed_message = kwargs["passed_message"]
    check.failed_message = kwargs["failed_message"]

    if must_contain and not test_result or cannot_contain and test_result:
        kwargs["test_status"] = TestResult.FAIL.name
        check.error_message = error_msg
        check.check_status = TestResult.FAIL.name
    else:
        check.check_status = TestResult.PASS.name

    kwargs["checks"] = [check]
    return kwargs


def function_calls_print_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name

    prog = Program(program_name=kwargs["file_name"])
    f_name = kwargs["function_name"]
    f = Function(f_name, prog.get_syntax_tree())

    test_result, error_msg = f.function_calls_print(kwargs["failed_message"])
    must_contain = kwargs["contains_check"][0]
    cannot_contain = kwargs["contains_check"][1]

    check = Check()
    check.before_message = kwargs["before_message"]
    check.passed_message = kwargs["passed_message"]
    check.failed_message = kwargs["failed_message"]

    if must_contain and not test_result or cannot_contain and test_result:
        kwargs["test_status"] = TestResult.FAIL.name
        check.error_message = error_msg
        check.check_status = TestResult.FAIL.name
    else:
        check.check_status = TestResult.PASS.name

    kwargs["checks"] = [check]
    return kwargs


def function_defines_function_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name
    prog = Program(program_name=kwargs["file_name"])
    f_name = kwargs["function_name"]
    f = Function(f_name, prog.get_syntax_tree())

    for check in kwargs["standard_output_checks"]:
        test_result, error_msg = f.function_defines_function2(check.string_check_type, check.expected_output,
                                                              check.nothing_else, check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
            break
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] = kwargs["standard_output_checks"]
    return kwargs


def function_imports_module_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name
    prog = Program(program_name=kwargs["file_name"])
    f_name = kwargs["function_name"]
    f = Function(f_name, prog.get_syntax_tree())

    for check in kwargs["standard_output_checks"]:
        test_result, error_msg = f.function_imports_module2(check.string_check_type, check.expected_output,
                                                            check.nothing_else, check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] = kwargs["standard_output_checks"]
    return kwargs


def function_contains_try_except_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name

    prog = Program(program_name=kwargs["file_name"])
    f_name = kwargs["function_name"]
    f = Function(f_name, prog.get_syntax_tree())

    test_result, error_msg = f.function_contains_try_except(kwargs["failed_message"])
    must_contain = kwargs["contains_check"][0]
    cannot_contain = kwargs["contains_check"][1]

    check = Check()
    check.before_message = kwargs["before_message"]
    check.passed_message = kwargs["passed_message"]
    check.failed_message = kwargs["failed_message"]

    if must_contain and not test_result or cannot_contain and test_result:
        kwargs["test_status"] = TestResult.FAIL.name
        check.error_message = error_msg
        check.check_status = TestResult.FAIL.name
    else:
        check.check_status = TestResult.PASS.name

    kwargs["checks"] = [check]
    return kwargs


def function_uses_only_local_vars_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name

    prog = Program(program_name=kwargs["file_name"])
    f_name = kwargs["function_name"]
    f = Function(f_name, prog.get_syntax_tree())

    test_result, error_msg = f.function_uses_only_local_vars(kwargs["failed_message"])

    must_contain = kwargs["contains_check"][0]
    cannot_contain = kwargs["contains_check"][1]

    check = Check()
    check.before_message = kwargs["before_message"]
    check.passed_message = kwargs["passed_message"]
    check.failed_message = kwargs["failed_message"]

    if must_contain and not test_result or cannot_contain and test_result:
        kwargs["test_status"] = TestResult.FAIL.name
        check.error_message = error_msg
        check.check_status = TestResult.FAIL.name
    else:
        check.check_status = TestResult.PASS.name

    kwargs["checks"] = [check]
    return kwargs


def function_execution_test(**kwargs):
    kwargs["test_status"] = TestResult.PASS.name
    kwargs["checks"] = []

    for file in kwargs["input_files"]:
        create_file(file[0], file[1])

    prog = Program(program_name=kwargs["file_name"])
    user_inputs_prog = kwargs["standard_input_data"]
    pe = ProgramExecution(prog, user_inputs_prog)
    f_name = kwargs["function_name"]

    try:
        pe.program_execution_and_input_count()
    except:
        kwargs["exception_message"] = pe.program_error_message
        kwargs["test_status"] = TestResult.FAIL.name
        return kwargs

    f_obj = pe.globals_dict.get(f_name, None)
    user_inputs_fun = kwargs["arguments"]
    fe = FunctionExecution(f_obj, user_inputs_fun, user_inputs_prog, kwargs["return_value"])

    try:
        execution_result, nr_of_user_inputs = fe.function_execution_and_input_count()
    except:
        pass

    check = Check()
    check.before_message = kwargs.get("before_message", "")
    check.passed_message = kwargs.get("passed_message", "")
    check.failed_message = kwargs.get("failed_message", "")
    test_result, error_msg = fe.function_return_value_correct()
    if test_result is False:
        kwargs["test_status"] = TestResult.FAIL.name
        check.error_message = error_msg
        check.check_status = TestResult.FAIL.name
        kwargs["checks"] += [check]
        return kwargs
    else:
        check.check_status = TestResult.PASS.name

    kwargs["checks"] += [check]

    kwargs["actual_output"] = fe.io_cap.get_stdout()

    # Standard output checks
    for check in kwargs["standard_output_checks"]:
        test_result, error_msg = fe.function_output_correct2(check.string_check_type, check.expected_output,
                                                                       check.nothing_else, kwargs["actual_output"],
                                                                       check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
            kwargs["checks"] += kwargs["standard_output_checks"]
            return kwargs
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] += kwargs["standard_output_checks"]

    # Output file output checks
    for check in kwargs["output_file_checks"]:
        output_file_name = check.file_name
        file_output = File(output_file_name).get_file_text()
        test_result, error_msg = prog.program_output_contains_keyword2(check.string_check_type, check.expected_output,
                                                                       check.nothing_else, file_output,
                                                                       check.failed_message)
        if not test_result:
            check.error_message = error_msg
            check.check_status = TestResult.FAIL.name
            kwargs["test_status"] = TestResult.FAIL.name
            kwargs["checks"] += kwargs["output_file_checks"]
            return kwargs
        else:
            check.check_status = TestResult.PASS.name

    kwargs["checks"] += (kwargs["output_file_checks"])
    return kwargs



import traceback
from math import ceil, floor
from re import findall
from typing import Any, List

from tiivad.file import File
from tiivad.results import Results
from tiivad.tests import program_imports_module_test, program_calls_function_test, program_contains_keyword_test, \
    program_contains_try_except_test, program_calls_print_test, program_contains_loop_test, program_execution_test, \
    function_execution_test, function_contains_loop_test, function_contains_keyword_test, function_contains_return_test, \
    function_calls_function_test, function_is_recursive_test, function_defines_function_test, \
    function_imports_module_test, function_contains_try_except_test, function_uses_only_local_vars_test, \
    function_calls_print_test


def validate_files(files: List[str]) -> bool:
    for file in files:
        f = File(file)

        f_exists, error_msg = f.file_exists()
        if not f_exists:
            Results.pre_evaluate_error = error_msg
            return False

        f_not_empty, error_msg = f.file_not_empty()
        if not f_not_empty:
            Results.pre_evaluate_error = error_msg
            return False

        f_is_python, error_msg = f.file_is_python()
        if not f_is_python:
            Results.pre_evaluate_error = error_msg
            return False

    return True


def execute_test(**kwargs):
    if Results.pre_evaluate_error:
        # print(Results(None).format_result())
        return

    try:
        if kwargs["type"] == "program_imports_module_test":
            Results(program_imports_module_test(**kwargs))
        elif kwargs["type"] == "program_calls_function_test":
            Results(program_calls_function_test(**kwargs))
        elif kwargs["type"] == "program_defines_function_test":
            Results(program_calls_function_test(**kwargs))
        elif kwargs["type"] == "program_contains_keyword_test":
            Results(program_contains_keyword_test(**kwargs))
        elif kwargs["type"] == "program_contains_try_except_test":
            Results(program_contains_try_except_test(**kwargs))
        elif kwargs["type"] == "program_calls_print_test":
            Results(program_calls_print_test(**kwargs))
        elif kwargs["type"] == "program_contains_loop_test":
            Results(program_contains_loop_test(**kwargs))
        elif kwargs["type"] == "program_execution_test":
            Results(program_execution_test(**kwargs))
        elif kwargs["type"] == "function_execution_test":
            Results(function_execution_test(**kwargs))
        elif kwargs["type"] == "function_contains_loop_test":
            Results(function_contains_loop_test(**kwargs))
        elif kwargs["type"] == "function_contains_keyword_test":
            Results(function_contains_keyword_test(**kwargs))
        elif kwargs["type"] == "function_contains_return_test":
            Results(function_contains_return_test(**kwargs))
        elif kwargs["type"] == "function_calls_function_test":
            Results(function_calls_function_test(**kwargs))
        elif kwargs["type"] == "function_is_recursive_test":
            Results(function_is_recursive_test(**kwargs))
        elif kwargs["type"] == "function_defines_function_test":
            Results(function_defines_function_test(**kwargs))
        elif kwargs["type"] == "function_imports_module_test":
            Results(function_imports_module_test(**kwargs))
        elif kwargs["type"] == "function_imports_module_test":
            Results(function_imports_module_test(**kwargs))
        elif kwargs["type"] == "function_contains_try_except_test":
            Results(function_contains_try_except_test(**kwargs))
        elif kwargs["type"] == "function_uses_only_local_vars_test":
            Results(function_uses_only_local_vars_test(**kwargs))
        elif kwargs["type"] == "function_calls_print_test":
            Results(function_calls_print_test(**kwargs))
        else:
            raise NotImplementedError("The following type of test is not implemented: " + str(kwargs["type"]))

    except Exception:
        Results.error = str(traceback.format_exc())


def execute_test_old(func, points, before_message, failed_message: str, passed_message, test_name, user_inputs, *args):
    if failed_message is None:
        failed_message = ""
    result = {}
    try:
        test_result, failed_message = func(*args, failed_message=failed_message)
    except Exception as e:  # TODO: Refactor me, too much duplication
        result['test_name_code'] = func.__name__
        result['points'] = 0.0
        result['exception_message'] = str(traceback.format_exc())
        result['before_message'] = before_message
        result['failed_message'] = str(e)
        result['passed_message'] = passed_message
        result['test_name'] = test_name
        result['user_inputs'] = user_inputs
        result['test_status'] = "Exception"
    else:
        if test_result:
            result['test_name_code'] = func.__name__
            result['points'] = points
            result['before_message'] = before_message
            result['failed_message'] = failed_message
            result['passed_message'] = passed_message
            result['test_name'] = test_name
            result['user_inputs'] = user_inputs
            result['test_status'] = "TestPass"
        else:
            result['test_name_code'] = func.__name__
            result['points'] = 0.0
            result['before_message'] = before_message
            result['failed_message'] = failed_message
            result['passed_message'] = passed_message
            result['test_name'] = test_name
            result['user_inputs'] = user_inputs
            result['test_status'] = "TestFail"
    Results(result)
    # print(result)
    return result


def create_input_file(file_name: str, file_content: str):
    f = open(file_name, "w", encoding="utf-8")
    f.write(file_content)
    f.close()


def get_file_text(file_name: str) -> str:
    with open(file_name, encoding='utf-8') as f:
        return f.read()


def value_not_empty(value: Any) -> bool:
    if value is None:
        return False
    if type(value) == str and value.strip() == "":
        return False
    return True


def string_not_empty(string: str) -> bool:
    return string is not None and string.strip() != ""


def unified_value_correct(value: Any, expected_value: Any, value_type: str = 'str') -> bool:
    # TODO lisa mõned testid
    # TODO: See on utils?
    """
    # * mida tagastatud väärtusele rakendatakse
    # * enne kui väärtuse õigsust kontrollitakse
    # - probleemid siis, kui programm tagastab ujukomaarve, seetõttu võiks vastust kontrollida hoopis vahemikku sobivusega
    # - funktsioon tagastab sõne, aga pole tegelikult oluline, et kas tegemist on suur- või väiketähtedega
    # - ehk me rakendame tudengi kirjutatud funktsiooni poolt tagastatud väärtusele mingi funktsiooni (nt .lower()) ja alles siis võrdleme etteantud tulemusega
    :param expected_value:
    :param result_value:
    :return:
    """
    if value_type == 'str':
        return value.strip().lower() == expected_value
    elif value_type == 'number':
        return floor(float(expected_value)) <= value <= ceil(float(expected_value))
    # TODO: Kas veel mõni tüüp?


def list_contains_items_ordered(lst: list, items: list) -> bool:
    """
    Abifunktsioon järgnevate jaoks
    """
    j = 0
    for item in items:
        while j < len(lst) and lst[j] != item:
            j += 1
        if j == len(lst):
            return False
        else:
            j += 1
    return True


# -- Numbers

def list_of_numbers(string: str) -> list[float]:
    if string is None:
        return []
    exp = r"([-]?[0-9]+(?:[.][0-9]+)?)"
    return list(map(float, findall(exp, string)))


def number_of_numbers(string: str) -> int:
    return len(list_of_numbers(string))


def number_of_numbers_exactly(string: str, n: int) -> bool:
    return number_of_numbers(string) == n


def number_of_numbers_at_least(string: str, n: int) -> bool:
    return number_of_numbers(string) >= n


def number_of_numbers_at_most(string: str, n: int) -> bool:
    return number_of_numbers(string) <= n


def string_contains_any_number(string: str) -> bool:
    return number_of_numbers_at_least(string, 1)


def string_contains_number(string: str, item: float) -> bool:
    return item in list_of_numbers(string)


def string_contains_numbers(string: str, items: list[float]) -> bool:
    sorted_numbers = sorted(list_of_numbers(string))
    sorted_items = sorted(items)
    return list_contains_items_ordered(sorted_numbers, sorted_items)


def string_contains_numbers_ordered(string: str, items: list[float]) -> bool:
    numbers = list_of_numbers(string)
    return list_contains_items_ordered(numbers, items)


def string_contains_only_these_numbers(string: str, items: list[float]) -> bool:
    sorted_numbers = sorted(list_of_numbers(string))
    sorted_items = sorted(items)
    return list_contains_items_ordered(sorted_numbers, sorted_items) and \
           len(sorted_numbers) == len(sorted_items)


def string_contains_only_these_numbers_ordered(string: str, items: list[float]) -> bool:
    numbers = list_of_numbers(string)
    return list_contains_items_ordered(numbers, items) and len(numbers) == len(items)


# -- Strings

def list_of_strings(string: str) -> list[str]:
    if string is None:
        return []
    s = string.replace(".", " ").replace(",", " ").replace(":", " ").replace(";", " ")
    s = s.replace("?", " ").replace("!", " ").replace('"', " ")
    return s.split()


def number_of_strings(string: str) -> int:
    return len(list_of_strings(string))


def number_of_strings_exactly(string: str, n: int) -> bool:
    return number_of_strings(string) == n


def number_of_strings_at_least(string: str, n: int) -> bool:
    return number_of_strings(string) >= n


def number_of_strings_at_most(string: str, n: int) -> bool:
    return number_of_strings(string) <= n


def string_contains_any_string(string: str) -> bool:
    return number_of_strings_at_least(string, 1)


def string_equals_string(string: str, item: str) -> bool:
    return string is not None and item == string


def string_contains_string(string: str, item: str) -> bool:
    return string is not None and item in string


def string_contains_strings(string: str, items: list[float]) -> bool:
    sorted_strings = sorted(list_of_strings(string))
    sorted_items = sorted(items)
    return list_contains_items_ordered(sorted_strings, sorted_items)


def string_contains_strings_ordered(string: str, items: list[float]) -> bool:
    strings = list_of_strings(string)
    return list_contains_items_ordered(strings, items)


def string_contains_only_these_strings(string: str, items: list[float]) -> bool:
    sorted_strings = sorted(list_of_strings(string))
    sorted_items = sorted(items)
    return list_contains_items_ordered(sorted_strings, sorted_items) and \
           len(sorted_strings) == len(sorted_items)


def string_contains_only_these_strings_ordered(string: str, items: list[float]) -> bool:
    strings = list_of_strings(string)
    return list_contains_items_ordered(strings, items) and len(strings) == len(items)


# -- Lines

def list_of_lines(string: str):
    if string is None:
        return []
    lines = []
    for s in string.strip().split("\n"):
        if s.strip() != "":
            lines.append(s)
    return lines


def number_of_lines(string: str) -> int:
    return len(list_of_lines(string))


def number_of_lines_exactly(string: str, n: int) -> bool:
    return number_of_lines(string) == n


def number_of_lines_at_least(string: str, n: int) -> bool:
    return number_of_lines(string) >= n


def number_of_lines_at_most(string: str, n: int) -> bool:
    return number_of_lines(string) <= n


def string_line_contains_string(string: str, line_nr: int, item: str) -> bool:
    lines = list_of_lines(string)
    if 1 <= line_nr <= len(lines):
        if item in lines[line_nr - 1]:
            return True
    return False


def string_line_equals(string: str, line_nr: int, item: str) -> bool:
    lines = list_of_lines(string)
    if 1 <= line_nr <= len(lines):
        if item == lines[line_nr - 1]:
            return True
    return False


def string_contains_lines(string: str, items: list[float]) -> bool:
    sorted_lines = sorted(list_of_lines(string))
    sorted_items = sorted(items)
    return list_contains_items_ordered(sorted_lines, sorted_items)


def string_contains_lines_ordered(string: str, items: list[float]) -> bool:
    lines = list_of_lines(string)
    return list_contains_items_ordered(lines, items)


def string_contains_only_these_lines(string: str, items: list[float]) -> bool:
    sorted_lines = sorted(list_of_lines(string))
    sorted_items = sorted(items)
    return list_contains_items_ordered(sorted_lines, sorted_items) and \
           len(sorted_lines) == len(sorted_items)


def string_contains_only_these_lines_ordered(string: str, items: list[float]) -> bool:
    lines = list_of_lines(string)
    return list_contains_items_ordered(lines, items) and len(lines) == len(items)



import os
import py_compile
from typing import Tuple

FILE_DOES_NOT_EXIST_MSG = "Faili nimega {file_name} ei eksisteeri."
FILE_IS_EMPTY_MSG = "Fail nimega {file_name} on tühi."
FILE_IS_NOT_VALID_PYTHON_MSG = "Fail nimega {file_name} ei ole korrektne Pythoni fail."


class File:
    def __init__(self, file_name: str):
        self.file_name: str = file_name
        self.content: str = None
        self.exists: bool = None
        self.not_empty: bool = None
        self.is_python: bool = None

    def file_exists(self, failed_message: str = FILE_DOES_NOT_EXIST_MSG) -> Tuple[bool, str]:
        """
        Checks if the file exists in the current working directory.
        :param failed_message: The message that will be displayed in case of failing the test.
        :return: True if file found, False if file not found.
        """
        if self.exists is None:
            self.exists = os.path.isfile(self.file_name)

        if not self.exists:
            failed_message = failed_message.format(file_name=self.file_name)

        return self.exists, failed_message

    def file_not_empty(self, failed_message: str = FILE_IS_EMPTY_MSG) -> Tuple[bool, str]:
        """
        Checks if the file is empty or not.
        :param failed_message: The message that will be displayed in case of failing the test.
        :return: True if file NOT empty, False if file is empty.
        """
        if not self.exists:
            return False, failed_message

        if self.not_empty is None:
            self.not_empty = os.stat(self.file_name).st_size > 0

        if not self.not_empty:
            failed_message = failed_message.format(file_name=self.file_name)

        return self.not_empty, failed_message

    def file_is_python(self, failed_message: str = FILE_IS_NOT_VALID_PYTHON_MSG) -> Tuple[bool, str]:
        """
        Checks if file is syntactically valid Python file.
        :param failed_message: The message that will be displayed in case of failing the test.
        :return: True if valid Python file, False if not.
        """
        if not self.exists:
            return False, failed_message

        if self.is_python is None:
            is_valid = True
            try:
                if not self.file_name.endswith(".py"):
                    is_valid = False
                py_compile.compile(self.file_name, doraise=True)
            except (SyntaxError, py_compile.PyCompileError) as e:
                is_valid = False
                failed_message = str(e)
            self.is_python = is_valid

        if not self.is_python:
            failed_message = failed_message.format(file_name=self.file_name)

        return self.is_python, failed_message

    def get_file_text(self) -> str:
        """
        Read in the content of the file.
        :return: str: Content of the file.
        """
        if not self.file_exists()[0]:
            return ""

        if self.content is None:
            with open(self.file_name, encoding='utf-8') as f:
                self.content = f.read()
        return self.content


if __name__ == '__main__':
    f = File("../test/samples/solution.py")
    print(f.file_exists())
    print(f.file_not_empty())
    print(f.file_is_python())
    a = f.get_file_text()
    print(a)



# from tiivad.file import File
# from tiivad.function import Function
# from tiivad.function_execution import FunctionExecution
# from tiivad.program_execution import ProgramExecution
# from tiivad.utils import create_input_file, execute_test, validate_files
import json

from tiivad.help_structures import StandardOutputChecks, OutputFileChecks
from tiivad.results import Results
#
# __all__ = ["File", "file", "function", "function_execution", "program", "program_execution", "utils",
#            "execute_test", "ProgramExecution", "Function", "create_input_file", "FunctionExecution", "Results",
#            "validate_files"]
from tiivad.utils import execute_test, validate_files


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'repr_json'):
            return obj.repr_json()
        else:
            return json.JSONEncoder.default(self, obj)


__all__ = ["execute_test", "validate_files", "StandardOutputChecks", "OutputFileChecks", "Results", "json",
           "ComplexEncoder"]



def seosta_lapsed_ja_vanemad(lapsed, nimed):
    f = open(lapsed, encoding='UTF-8')
    ff = open(nimed, encoding='UTF-8')
    koodid = []
    for i in f:
        i = i.strip().split()
        koodid.append(i)

    vl = []
    lapsed = set()
    for i in koodid:
        ff.seek(0, 0)
        for n in ff:
            n = n.split()
            if i[0] in n:
                v = n[1] + ' ' + n[2]
            elif i[1] in n:
                l = n[1] + ' ' + n[2]
                lapsed.add(l)
        nim = v, l
        vl.append(nim)

    f.close()
    ff.close()

    lapsevanemad = {}
    for i in lapsed:
        emaisa = set()
        for v in vl:
            if i in v:
                if i != v[0]:
                    emaisa.add(v[0])
        lapsevanemad[i] = emaisa

    return lapsevanemad


nimed = seosta_lapsed_ja_vanemad('lapsed.txt', 'nimed.txt')
lapsed = list(nimed.keys())

for i in lapsed:
    vanemad = ''
    for n in nimed[i]:
        vanemad += n + ' '
    print(i + ':', vanemad)



import os
import traceback
from typing import Tuple

from tiivad.edutest import run_script, OutOfInputsError, IOCapturing
from tiivad.program import Program
from tiivad.t import format_error


class ProgramExecution:
    def __init__(self, program: Program, user_inputs: list = [], expected_output: str = ""):
        self.program = program
        self.user_inputs: list = user_inputs
        self.globals_dict: dict = None
        self.nr_of_user_inputs: int = None
        self.program_threw_error: bool = None
        self.io_cap: IOCapturing = None
        self.expected_output: str = expected_output
        self.program_exists: bool = None
        self.is_python_file: bool = None  # TODO: Fill me in - samas kohas kus kontrollitakse, et kas fail on tühi.
        self.program_error_message: str = None

    def program_execution_and_input_count(self, re_throw_exception: bool = True, failed_message: str = ""):
        """
        # TODO: What to do with failed_message here?
        :return:
        """
        self.program_threw_error = False
        self.globals_dict = {}
        try:
            if self.user_inputs is None:
                self.user_inputs = []
            if not self.program.file.file_exists()[0]:
                self.program_exists = False  # TODO: Think if we need this variable at all.
                return False, ""
            globs, cap = run_script(self.program.file.file_name, self.user_inputs)
            self.io_cap = cap
            self.globals_dict = globs
        except OutOfInputsError:
            self.nr_of_user_inputs = float('Inf')
            self.program_threw_error = True
            self.program_error_message = "Programm küsib rohkem sisendeid kui oli ette antud."
            return False, self.program_error_message
        except Exception as e:
            self.program_threw_error = True
            if re_throw_exception:
                self.program_error_message = str(traceback.format_exc())
                raise e
            return True, ""
        else:
            self.nr_of_user_inputs = len(self.user_inputs) - len(cap.get_remaining_inputs())
            return True, ""

    def program_input_count_correct(self, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if the number of user inputs asked by the program is correct.
        :return: True if expected nr of user inputs matches actual count of user inputs, False if not.
        """
        if self.program_threw_error:
            return False, self.program_error_message
        if self.nr_of_user_inputs is None:
            self.program_execution_and_input_count()
        return self.nr_of_user_inputs == len(self.user_inputs), format_error(failed_message, {
            "nr_of_user_inputs": self.nr_of_user_inputs, "user_inputs": self.user_inputs})

    def program_input_count_more_than_expected(self):
        """
        Checks if the number of user inputs asked by the program is more than provided (can be infinite).
        :return: True if actual number of user inputs asked is bigger than expected, False if not.
        """
        if self.program_threw_error:
            return False, self.program_error_message
        if self.nr_of_user_inputs is None:
            self.program_execution_and_input_count()
        return self.nr_of_user_inputs == float('Inf')

    def program_raised_exception(self, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks whether the program threw an exception or not.
        :return: True if threw, False if not.
        """
        if self.program_threw_error is None:
            self.program_execution_and_input_count()
        return self.program_threw_error, failed_message

    def program_output_correct(self, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if the program output is correct.
        :return: True if program output is correct, False if not.
        """
        if self.program_threw_error:
            return False, self.program_error_message
        if not self.io_cap:
            return False, failed_message
        return str(self.expected_output) in self.io_cap.get_stdout(), failed_message

    def program_output_file_correct(self, file_name: str, file_content: str, failed_message: str = "") -> Tuple[
        bool, str]:
        """
        Checks if the output file is correct.
        :return: True if program output file is correct, False if not.
        """
        if self.program_threw_error:
            return False, self.program_error_message
        if os.path.isfile(file_name):
            with open(file_name) as f:
                lines = f.read()
                return str(lines) in file_content, failed_message
        else:
            return False, failed_message


if __name__ == '__main__':
    prog = Program("../test/samples/solution.py")
    user_inputs = [5, 3] * 3
    pe = ProgramExecution(prog, user_inputs)
    pe.program_execution_and_input_count()
    print(pe.nr_of_user_inputs)
    print(pe.program_input_count_correct())
    print(pe.program_input_count_more_than_expected())
    print(pe.program_raised_exception())
    print("IOCap output:")
    print(pe.io_cap.get_stdout())
    print("Last output:")
    print(pe.io_cap.get_last_stdout())
    print(pe.program_output_correct("user input 1"))



import os.path
from inspect import signature
from math import ceil, floor
from typing import Callable, Union

from edutest import *


# File checks
def file_exists(file_name: str) -> bool:
    """
    Checks if the file exists in the current working directory.
    :param file_name: str: File name to be checked.
    :return: True if file found, False if file not found.
    """
    return os.path.isfile(file_name)


def file_not_empty(file_name: str) -> bool:
    """
    Checks if the file is empty or not.
    :param file_name: str: File name to be checked.
    :return: True if file NOT empty, False if is empty.
    """
    return os.stat(file_name).st_size > 0


# Täidab programmi etteantud sisenditega
# TODO: Refactor, et teeks vaid ühte asja.
def file_is_python(programminimi, sisendid) -> Union[dict, IOCapturing, float]:
    try:
        globs, cap = run_script(programminimi, sisendid)
    except OutOfInputsError:
        tegelik_sisendite_arv = float('Inf')
    else:
        tegelik_sisendite_arv = len(sisendid) - len(cap.get_remaining_inputs())

    return globs, cap, tegelik_sisendite_arv


# Täidab funktsiooni etteantud argumentide ja sisenditega, kontrollib ka sisendite arvu.
def täidafunk(funktsioon, argumendid, sisendid, oodatav_sisendite_arv=0, veatekst=""):
    try:
        with IOCapturing(sisendid) as cap:
            tegelik_tulemus = funktsioon(*argumendid)
    except Exception as e:
        print(e)
        raise
    return tegelik_tulemus, cap


def fun_returned(returned_result):
    return returned_result is not None


def fun_prints(function):
    return fun_uses_fun(function, "print")


# TODO: See on ainult veateate kirjutamiseks, tuleks mõelda, et kuidas neid üleüldse hallata (veateateid)
# Programmi sisendparameetrite arvu kontrollimine ja teadete väljastamine.
# Veatekst on täiendav info kasutajale vea korral.
def prog_input_count(tegelik_sisendite_arv, oodatav_sisendite_arv, veatekst=""):
    if tegelik_sisendite_arv != oodatav_sisendite_arv:
        if tegelik_sisendite_arv == 0 and oodatav_sisendite_arv > 0:
            teade = "Programm ei küsinud sisendeid.\n"
            teade += "Programm peaks küsima sisendit kasutajalt."
            print(teade)
        elif tegelik_sisendite_arv > 0 and oodatav_sisendite_arv == 0:
            teade = "Programm küsis sisendit.\n"
            teade += "Programm ei tohiks kasutajalt ühtegi sisendit küsida."
            print(teade)
        elif float('Inf') > tegelik_sisendite_arv > 0 and oodatav_sisendite_arv > 0:
            teade = "Programm küsis "
            teade += ("vähem" if tegelik_sisendite_arv < oodatav_sisendite_arv else "rohkem")
            teade += " sisendeid, kui oleks pidanud.\n"
            teade += "Programm küsis " \
                     + ", aga oleks pidanud küsima " + ".\n"
            teade += veatekst
            print(teade)
        elif tegelik_sisendite_arv == float('Inf') and oodatav_sisendite_arv > 0:
            teade = "Programm küsis rohkem sisendeid, kui oleks pidanud.\n"
            teade += "Programm peaks küsima ainult " + ".\n"
            teade += veatekst
            print(teade)


# Function checks:
def prog_has_fun(globals_dict: dict, fun_name: str) -> bool:
    # Programm sisaldab antud funktsiooni definitsiooni
    return fun_name in globals_dict and inspect.isfunction(globals_dict[fun_name])


def fun_input_count_correct(function: Callable, expected_no_of_params: int) -> bool:
    # Funktsioonil on õige arv parameetreid
    # TODO: Nimetada ümber? Sest see nimetus viitaks, et tagastataks arv, mitte tõeväärtus
    return len(signature(function).parameters) == expected_no_of_params


def check_fun(function: Callable, *args):
    # funktsiooni definitsioon on süntaktiliselt korrektne
    # - standardteek, mis kontrolliks, kas etteantud tekst on valiidne Pythoni programm või mitte
    # - kui file_is_python tagastab veateate, siis see funktsioon võib ikkagi tagastada tõese väärtuse.
    # Kontrollime, et kas funktsioon läheb käima.
    try:
        function(*args)
        return True
    except:
        return False


def fun_input_type(function: Callable, signature_types: list) -> bool:
    # * funktsiooni parameetrid on õiget tüüpi
    # TODO: Kuidas me seda kontrollida võiksime?
    # [param_type_1, param_type_2, param_type_n]
    sig = signature(function)
    p = signature(function).parameters
    # return len() == expected_no_of_params
    # TODO: Pythoni puhul jätame vahele!
    pass


def fun_contains_return(fun_name: Callable):
    # * funktsiooni definitsioon sisaldab etteantud võtmesõna
    # TODO: Miks meil seda vaja on kui me juba teame, et vastav funktsioon eksisteerib?
    # nt kontrollib, et kas tsükli käsk eksisteerib
    # tihti aetakse sassi return ja print
    # Üldisem - "text" ei pea olema funktsioon
    for node in ast.walk(ast.parse(inspect.getsource(fun_name))):
        if isinstance(node, ast.Return):
            return True
    return False


def fun_contains_loop(fun_name: Callable):
    for node in ast.walk(ast.parse(inspect.getsource(fun_name))):
        if isinstance(node, (ast.For, ast.While)):
            return True
    return False


def fun_uses_fun(function: Callable, contains_fun: str) -> bool:
    # * funktsioon kasutab etteantud funktsiooni
    # - nt rekursiivne väljakutse
    # saab kasutada progkontrolli funkpöördub nt
    # Vahel on tarvis kontrollida mitme funktsiooni olemasolu funktsioonis/programmis. Äkki on optimaalsem süntaksipuu
    # luua ühe korra? Võimalus teha funktsioon, mis tagastab süntaksipuu ja see antakse teistele funktsioonidele
    # sisendiks.
    return contains_fun in [c.func.id for c in ast.walk(ast.parse(inspect.getsource(function))) if
                            isinstance(c, ast.Call)]


def fun_uses_local_vars(fun_name: str):
    # * funktsioon kasutab ainult funktsiooni sees defineeritud muutujaid
    # - tarvis kindlaks teha terve programmi globaalsed muutujad
    # - seejärel teha kindlaks, kas kontrollitav funktsioon kasutab sama nimega mõnda muutujat, kui jah, siis vaadata,
    # Implementatsioon: Kas funktsioon kasutab mõnda muutujat, mis ei ole eelnevalt defineeritud selle funktsiooni sees?
    # Võtmesõna "global" kasutamine ka keelata ära
    # 1. Kontrollida, et kas programmi tekst sisaldab sõna "global"
    # 2. Teha kindlaks kõigi programmis defineeritud muutujate hulk
    # 3. Kontrollida, et kui funktsiooni sees oleva programmi muutuja väärtust on enne funktsiooni muudetud, siis on viga.
    defined_vars = set()
    tree = ast.parse(inspect.getsource(fun_name))
    for node in ast.walk(tree):
        if isinstance(node, ast.arg):
            defined_vars.add(node.arg)
            if node.annotation != None:
                defined_vars.add(node.annotation.id)
        if isinstance(node, ast.Name) and isinstance(node.ctx, ast.Store):
            defined_vars.add(node.id)
        if isinstance(node, ast.Call):
            defined_vars.add(node.func.id)
        if isinstance(node, ast.Name) and isinstance(node.ctx, ast.Load):
            if node.id not in defined_vars:
                return False
    return True


def fun_user_input_count(fun_name: str, no_of_inputs: int):
    # Funktsioon küsib kasutajalt õiget arvu sisendeid
    pass
    # TODO: Kas seda ei tee juba funktsioon 'täidafunk'?


def fun_modified_param_type(actual_solution, example_solution):
    # * funktsiooni parameetri muudetud väärtuse tüübistruktuur on õige
    # - nt tuleb etteantud parameetrit muuta (nt järjend), aga funktsioon ise ei tagasta midagi
    # - lõppväärtuse juures tuleb kontrollida, et kas muudetud parameeter on õigete tüüpidega
    # - Kas lihtsalt "jooksutan" funktsiooni ja seejärel vaatan muutuja tüüpi?
    if isinstance(actual_solution, (list, tuple)):
        if len(actual_solution) != len(example_solution):
            return False
        for i in range(len(actual_solution)):
            if not fun_modified_param_type(actual_solution[i], example_solution[i]):
                return False
        return True
    else:
        return type(actual_solution) == type(example_solution)


def fun_modified_param_value(actual_solution, example_solution):
    # funktsiooni parameetri muudetud väärtus on õige
    return actual_solution == example_solution


def check_result(result, expected_result, result_type='str'):
    """
    # * mida tagastatud väärtusele rakendatakse
    # * enne kui väärtuse õigsust kontrollitakse
    # - probleemid siis, kui programm tagastab ujukomaarve, seetõttu võiks vastust kontrollida hoopis vahemikku sobivusega
    # - funktsioon tagastab sõne, aga pole tegelikult oluline, et kas tegemist on suur- või väiketähtedega
    # - ehk me rakendame tudengi kirjutatud funktsiooni poolt tagastatud väärtusele mingi funktsiooni (nt .lower()) ja alles siis võrdleme etteantud tulemusega
    :param result:
    :param expected_result:
    :param result_type:
    :return:
    """
    if result_type == 'str':
        return result.strip().lower() == expected_result
    elif result_type == 'number':
        return floor(float(expected_result)) <= result <= ceil(float(expected_result))
    # TODO: Kas veel mõni tüüp?


def fun_prints_instead_of_returning(fun_name, actual_result):
    return not fun_returned(actual_result) and fun_prints(fun_name)


def fun_return_type_correct(returned_value, expected_type):
    return type(returned_value) == expected_type


def fun_return_value_correct(returned_value, expected_value):
    return returned_value == expected_value


# Väljundi kontrollid
def prog_output_not_empty(output):
    return output is not None and output.strip() != ""


def prog_output_contains_string(output, str):
    return output is not None and str in output


def prog_output_contains_items(output, items):
    if output is None:
        return False
    for i in items:
        if i not in output:
            return False
    return True


def prog_output_contains_only_these_items(output, items):
    if output is None:
        return False
    return sorted(output.strip().split("\n")) == sorted(items)


def prog_output_contains_ordered(output, items):
    if output is None:
        return False
    return output.strip().split("\n") == items


def prog_output_no_of_lines(output, no_of_lines: int):
    if output is None:
        return False
    # TODO: Kas siin peaks tühjad read eemaldama või mitte?
    return len(output.strip().split("\n")) == no_of_lines


def prog_output_contains_no_of_elements_of_type(output, el_type, no_of_elements):
    if output is None:
        return False
    formatted_output = output.strip().split("\n")
    c = 0
    for i in formatted_output:
        if c > no_of_elements:
            return False
        if type(i) == el_type:
            c += 1
    return c == no_of_elements


def prog_output_no_of_lines_not_empty(output, no_of_lines: int):
    if output is None:
        return False
    formatted_output = output.strip().split("\n")
    return len(formatted_output) == no_of_lines


def prog_output_line_contains(output, item, line_nr, arvesta_tühjade_ridadega=False):
    if output is None:
        return False
    formatted_output = output.strip().split("\n")
    return item in formatted_output[line_nr - 1]


def prog_output_line_equals(output, item, line_nr, arvesta_tühjade_ridadega=False):
    if output is None:
        return False
    formatted_output = output.strip().split("\n")
    return item == formatted_output[line_nr - 1]


def prog_output_ordered(output, items, arvesta_tühjade_ridadega=False):
    if output is None:
        return False
    # TODO: Kas see pole sama mis üks eelmine funktsioon?
    formatted_output = output.strip().split("\n")
    return formatted_output == items


def prog_output_contains(output, items, arvesta_tühjade_ridadega=False):
    if output is None:
        return False
    # TODO: Kas see pole sama mis üks eelmine funktsioon?
    return sorted(output.strip().split("\n")) == sorted(items)


def get_prog_text(file_name):
    with open(file_name, encoding='utf-8') as f:
        return f.read()


def prog_contains_text(prog_text, text: str):
    if prog_text is None:
        return False
    if text in prog_text:
        return True
    return False


def output_text_correct(output_text, text):
    if output_text is None:
        return False
    if text == output_text:
        return True
    return False


# Kontrollib, kas programm kasutab moodulit
def prog_uses_module(programmisisu, moodulinimi):
    # TODO: Refactor
    # with open(programminimi, encoding='utf-8') as f:
    #     programmisisu = f.read()
    # print(programmisisu)
    süntaksipuu = ast.parse(programmisisu)
    if type(moodulinimi) == str:
        kontrollitavad = [moodulinimi]
    elif type(moodulinimi) == list:
        kontrollitavad = moodulinimi
    for tipp in ast.walk(süntaksipuu):
        if isinstance(tipp, ast.Import):
            if 'names' in tipp._fields:
                for nimi in tipp.names:
                    if 'name' in nimi._fields:
                        for i in range(len(kontrollitavad)):
                            if nimi.name == kontrollitavad[i]:
                                return True
        if isinstance(tipp, ast.ImportFrom):
            if 'module' in tipp._fields:
                for i in range(len(kontrollitavad)):
                    if tipp.module == kontrollitavad[i]:
                        return True
    return False


def prog_uses_fun(programmisisu, funktsiooninimi):
    # TODO: Refactor
    süntaksipuu = ast.parse(programmisisu)
    if type(funktsiooninimi) == str:
        kontrollitavad = [funktsiooninimi]
    elif type(funktsiooninimi) == list:
        kontrollitavad = funktsiooninimi
    for tipp in ast.walk(süntaksipuu):
        if isinstance(tipp, ast.Call):
            if 'func' in tipp._fields:
                if 'id' in tipp.func._fields:
                    for i in range(len(kontrollitavad)):
                        if tipp.func.id == kontrollitavad[i]:
                            return True
                if 'attr' in tipp.func._fields:
                    for i in range(len(kontrollitavad)):
                        if tipp.func.attr == kontrollitavad[i]:
                            return True
    return False


def prog_uses_fun_from_set(program_data, functions_list):
    # TODO: Eelmine funktsioon tegelikult juba teeb seda.
    pass


if __name__ == "__main__":
    print(file_exists("progkontroll.py"))
    print(file_not_empty("progkontroll.py"))
    sisendid = [5, 3] * 20
    oodatav_sisendite_arv = 2
    veatekst = "Programmi sisendid peaksid olema elektriliini pikkus ja kõrvutiasetsevate postide maksimaalkaugus."
    globals_dict, cap, tegelik_sisendite_arv = file_is_python("solution.py", sisendid)
    prog_input_count(tegelik_sisendite_arv, oodatav_sisendite_arv, veatekst)
    print(prog_has_fun(globals_dict, 'test'))
    print(globals_dict['test'])
    print(type(globals_dict['test']))
    # print(fun_input_count(globals_dict['test'], 1))
    print(fun_input_type(globals_dict['test'], [str, str, int]))
    print(fun_uses_fun(globals_dict['test'], 'nottest'))
    # print(globals_dict['test'].__dir__)
    # print(globals_dict.globals)

    print("R")
    prog_output = cap.get_last_stdout()
    print(prog_output)
    print("E")

    tegelik_tulemus, cap = täidafunk(globals_dict['test'], [1], "sisend")
    print(tegelik_tulemus, cap)
    print(fun_returned(tegelik_tulemus))
    print(fun_prints(globals_dict['test']))
    print(fun_prints_instead_of_returning(globals_dict['test'], tegelik_tulemus))
    print(fun_return_type_correct(tegelik_tulemus, int))
    print(fun_return_value_correct(tegelik_tulemus, "2"))

    fun_output = cap.get_last_stdout()
    print("E")

    print(prog_output_not_empty(prog_output))
    print(prog_output.strip())
    print(type(prog_output))
    print(prog_output_contains_string(prog_output, "liini"))
    print(prog_output_contains_string(prog_output, "Liini"))
    print(prog_output_contains_items(prog_output, ["Liini", "iini"]))
    print(
        prog_output_contains_only_these_items(prog_output, ['Liini ehitamiseks läheb vaja minimaalselt 3 posti.', 'u']))
    print(prog_output_contains_ordered(prog_output, ['u', 'Liini ehitamiseks läheb vaja minimaalselt 3 posti.']))
    print(prog_output_contains_ordered(prog_output, ['Liini ehitamiseks läheb vaja minimaalselt 3 posti.', 'u']))
    print(prog_output_no_of_lines(prog_output, 3))
    print(prog_output_no_of_lines(prog_output, 2))
    print(prog_output_contains_no_of_elements_of_type(prog_output, int, 2))
    print(prog_output_contains_no_of_elements_of_type(prog_output, str, 2))
    print(prog_output_contains_no_of_elements_of_type(prog_output, str, 1))
    print(prog_output_contains_no_of_elements_of_type(prog_output, str, 3))
    print(prog_output_no_of_lines_not_empty(prog_output, 2))
    print(prog_output_no_of_lines_not_empty(prog_output, 1))
    print(prog_output_line_contains(prog_output, "ehita", 1))
    print(prog_output_line_contains(prog_output, "ehita", 2))
    print(prog_output_line_contains(prog_output, "asd", 1))
    print(prog_output_line_equals(prog_output, "u", 2))
    print(prog_output_line_equals(prog_output, "u", 1))
    print(prog_output_ordered(prog_output, ['Liini ehitamiseks läheb vaja minimaalselt 3 posti.', 'u']))
    print(prog_output_ordered(prog_output, ['u', 'Liini ehitamiseks läheb vaja minimaalselt 3 posti.']))
    print(prog_output_contains(prog_output, ['Liini ehitamiseks läheb vaja minimaalselt 3 posti.', 'u']))
    print(prog_output_contains(prog_output, ['u', 'Liini ehitamiseks läheb vaja minimaalselt 3 posti.', 'u']))
    print(prog_output_contains(prog_output, ['u', 'Liini ehitamiseks läheb vaja minimaalselt 3 posti.']))

    prog_text = get_prog_text('../test/samples/solution.py')
    print(prog_contains_text(prog_text, "postide"))
    print(prog_contains_text(prog_text, "postide2"))
    print(prog_uses_module(prog_text, "turtle"))
    print(prog_uses_module(prog_text, "math"))
    print(prog_uses_module(prog_text, "ceil"))
    print(prog_uses_fun(prog_text, "test"))
    print(prog_uses_fun(prog_text, "test2"))

    print(check_result(1, 1.2, 'number'))
    print(check_result(2, 1.2, 'number'))
    print(check_result(1.5, 1.5, 'number'))
    print(check_result(0.9, 1.5, 'number'))
    print(check_result(2.1, 1.5, 'number'))

    print(check_result('asd ', 'asd', 'str'))
    print(check_result('asd s', 'asd', 'str'))
    print(check_result('asD', 'asd', 'str'))

    print(fun_contains_return(globals_dict['test']))
    print(fun_contains_loop(globals_dict['test']))
    print(fun_uses_fun(globals_dict['test'], "print"))

    print(check_fun(globals_dict['test'], 2))



def format_error(error_message, mapping_dict):
    for k, v in mapping_dict.items():
        a = {k: "{" + str(v) + "}"}
        try:
            error_message = error_message.format(**a)
        except:
            pass
    return error_message



from typing import Any, Callable, Tuple, List

from tiivad.edutest import IOCapturing, OutOfInputsError
from tiivad.program import Program, ValidationType
from tiivad.program_execution import ProgramExecution


class FunctionExecution:
    def __init__(self, function_object: Callable, function_arguments: list, user_inputs: list, expected_result):
        self.execution_result = None
        self.nr_of_user_inputs: int = None
        self.function_executed: bool = False
        self.function_object: Callable = function_object
        self.user_inputs: list = [] if user_inputs is None else user_inputs
        self.function_threw_error: bool = None
        self.function_arguments: list = function_arguments
        self.io_cap: IOCapturing = None
        self.expected_result = expected_result if expected_result is not None else expected_result
        self.function_error_message: str = None

    def function_execution_and_input_count(self, failed_message: str = "", re_throw_exception: bool = True):
        """
        Execute function with the provided arguments and checks if number of user inputs matches expected.
        :param re_throw_exception: Boolean deciding whether to re-throw the error after catching.
        :return: Tuple consisting of execution result, IOCap and the actual number of user inputs.
        """
        self.function_executed = True
        self.function_threw_error = False
        try:
            if self.function_object is None:
                self.function_threw_error = True
                self.function_error_message = "Antud nimega funktsiooni ei ole defineeritud."
                return False, self.function_error_message
            with IOCapturing(self.user_inputs) as cap:
                self.execution_result = self.function_object(*self.function_arguments)
            self.io_cap = cap
        except OutOfInputsError:
            self.nr_of_user_inputs = float('Inf')
            return True, failed_message
        except Exception as e:
            self.function_threw_error = True
            self.function_error_message = e
            if re_throw_exception:
                raise e
            return False, failed_message
        else:
            self.nr_of_user_inputs = len(self.user_inputs) - len(cap.get_remaining_inputs())
            return True, failed_message

    def function_input_count_correct(self) -> bool:
        """
        Checks if the number of user inputs asked by the function is correct.
        :return: True if expected nr of user inputs matches actual count of user inputs, False if not.
        """
        if self.function_threw_error:
            return False, self.function_error_message
        return self.nr_of_user_inputs == len(self.user_inputs)

    def function_input_count_more_than_expected(self) -> bool:
        """
        Checks if the number of user inputs asked by the function is more than provided (can be infinite).
        :return: True if actual number of user inputs asked is bigger than expected, False if not.
        """
        if self.function_threw_error:
            return False, self.function_error_message
        return self.nr_of_user_inputs == float('Inf')

    def function_modified_param_type_correct(self, actual_solution, example_solution) -> bool:
        """
        # TODO: Rename input parameters.
        Checks if the parameter type is correct after being modified by the function.
        :param actual_solution: TODO:
        :param example_solution: TODO
        :return: True if the type is correct, False if not.
        """
        # * funktsiooni parameetri muudetud väärtuse tüübistruktuur on õige
        # - nt tuleb etteantud parameetrit muuta (nt järjend), aga funktsioon ise ei tagasta midagi
        # - lõppväärtuse juures tuleb kontrollida, et kas muudetud parameeter on õigete tüüpidega
        # - Kas lihtsalt "jooksutan" funktsiooni ja seejärel vaatan muutuja tüüpi?
        if self.function_threw_error:
            return False, self.function_error_message
        if isinstance(actual_solution, (list, tuple)):
            if len(actual_solution) != len(example_solution):
                return False
            for i in range(len(actual_solution)):
                if not self.function_modified_param_type_correct(actual_solution[i], example_solution[i]):
                    return False
            return True
        else:
            return type(actual_solution) == type(example_solution)

    def function_modified_param_value_correct(self, actual_solution, example_solution) -> bool:
        # TODO: Rename input parameters.
        """
        Checks if the parameter value is correct after being modified by the function.
        :param actual_solution: # TODO
        :param example_solution: # TODO
        :return:
        """
        if self.function_threw_error:
            return False, self.function_error_message
        return actual_solution == example_solution

    def function_returned(self, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks if the function got executed and if it actually returned anything.
        :return: True if function executed and returned something, False if function not executed or didn't return.
        """
        if self.function_threw_error:
            return False, self.function_error_message
        return self.function_executed and self.execution_result is not None, failed_message

    def function_return_type_correct(self, expected_value: str, failed_message: str = "") -> Tuple[bool, str]:
        """
        # TODO: See on utils?
        :param expected_value:
        :return:
        """
        if self.function_threw_error:
            return False, self.function_error_message
        return type(self.execution_result) == eval(expected_value), failed_message

    def function_return_value_correct(self, failed_message: str = "Funktsiooni eeldatav väljund pidi olema "
                                                                  "{expected_output}, kuid tulemus oli {output}") \
            -> Tuple[bool, str]:
        """

        :param failed_message:
        :return:
        """
        if self.function_threw_error:
            return False, self.function_error_message
        return self.execution_result == self.expected_result, failed_message.format(
            expected_output=str(self.expected_result), output=str(self.execution_result))

    def function_raised_exception(self, failed_message: str = "") -> Tuple[bool, str]:
        """
        Checks whether the function threw an exception or not.
        :return: True if threw, False if not.
        """
        return self.function_threw_error, failed_message

    @staticmethod
    def function_output_contains_keyword(function_output_str: str, keyword: str, failed_message: str = "") \
            -> Tuple[bool, str]:
        """

        :param function_output_str:
        :param keyword:
        :param failed_message:
        :return:
        """
        if keyword in function_output_str:
            return True, ""
        else:
            return False, failed_message

    def function_output_correct2(self, validation_type: str, outputs: List[str], nothing_else: bool, actual_output: str,
                                 failed_message: str = "") -> Tuple[bool, str]:
        """

        :param actual_output:
        :param validation_type:
        :param outputs:
        :param nothing_else:
        :param failed_message:
        :return:
        """
        if self.function_threw_error:
            return False, self.function_error_message

        if validation_type == ValidationType.ALL_OF_THESE.name:
            for func in outputs:
                if not self.function_output_contains_keyword(actual_output, func, failed_message)[0]:
                    return False, failed_message

        elif validation_type == ValidationType.ANY_OF_THESE.name:
            for func in outputs:
                if self.function_output_contains_keyword(actual_output, func, failed_message)[0]:
                    return True, ""
            return False, failed_message

        elif validation_type == ValidationType.NONE_OF_THESE.name:
            for func in outputs:
                if self.function_output_contains_keyword(actual_output, func, failed_message):
                    return False, failed_message

        elif validation_type == ValidationType.MISSING_AT_LEAST_ONE_OF_THESE.name:
            for func in outputs:
                if not self.function_output_contains_keyword(actual_output, func, failed_message)[0]:
                    return True, failed_message
            return False, failed_message

        else:
            raise NotImplementedError(
                f"The following type is not implemented for the program_calls_function_test: {validation_type}")

        return True, ""


if __name__ == '__main__':
    prog = Program("../test/samples/solution.py")
    user_inputs_prog = [5, 3] * 3
    p = ProgramExecution(prog, user_inputs_prog)
    f_name = "test"
    p.program_execution_and_input_count()
    f_obj = p.globals_dict[f_name]
    user_inputs_fun = [5, 3] * 2
    fe = FunctionExecution(f_obj, [1], user_inputs_fun)
    execution_result, nr_of_user_inputs = fe.function_execution_and_input_count()
    print(fe.function_returned())
    print(fe.function_input_count_correct())
    print(fe.function_input_count_more_than_expected())
    print(fe.function_return_type_correct(int))
    print(fe.function_return_value_correct(2))
    print(fe.function_raised_exception())
    print("IOCap output:")
    print(fe.io_cap.get_stdout())



import re
import ast
import sys
import runpy
import types
import locale
import inspect
import traceback
import importlib
import os.path
from copy import deepcopy

OUT_OF_INPUTS = "Program asked for input but no more input was available"
INPUT_NOT_ALLOWED = "Reading input is not allowed"
OUTPUT_NOT_ALLOWED = "Writing output is not allowed"
CANT_FIND_FUNCTION = "Can't find function {name}"
CANT_FIND_FILE = "Can't find file {path}"
CANT_IMPORT_MODULE_WITH_INPUT = "Can't import {file} as module, because it tried to read input"
FUNCTION_MODIFIED_ARGUMENTS = "Your function modified its argument objects"
FUNCTION_MODIFIED_GLOBALS = "Your function modified global variables or objects"
FUNCTION_PRINTED_INSTEAD_OF_RETURNING = "Function printed correct response but should have returned it"
FUNCTION_RETURNED_WRONG_TYPE = "Function returned a value of type {actual}, but should have returned a value of type {expected}"
FUNCTION_RETURNED_WRONG_VALUE = "Function returned {actual}, but should have returned {expected}"
FUNCTION_READ_STDIN = "Function tried to read input"
FUNCTION_WROTE_STDOUT = "Function printed something but this was not allowed"
FUNCTION_WROTE_STDERR = "Function wrote to error stream but this was not allowed"
CANT_FIND_TARGET = "Can't find the script to test"

ALLOW_DECIMAL_COMMA = False

_target = None
_test_file_prefixes = ["test_"]
_test_file_suffixes = ["_test", "_tests", "_tester"]
TARGET_SCRIPT_ENV_KEY = "TESTABLE_SCRIPT"


def set_target(path):
    """Fixes a file name to be used when checking functions are called without file name argument."""
    global _target
    _target = path


def find_target():
    """If testable script path has been fixed with :func:`set_target` then
    returns this file.

    Otherwise, if there is environment variable TARGET_SCRIPT_ENV_KEY set, then its
    value is returned.

    Otherwise the inference algorithm tries to detect current test file and the testable script
    with matching name. It succeeds if testable script and test file are both in
    current directory and test file name is formed by appending word ``_test``, ``_tests``
    or ``_tester`` to script name, or prepending ``test_`` to script name.

    Example: "exercise1.py" is  found if test file is one of

    * exercise1_test.py
    * exercise1_tests.py
    * exercise1_tester.py
    * test_exercise1.py

    """
    if _target is not None:
        return _target
    else:
        if (TARGET_SCRIPT_ENV_KEY in os.environ
                and os.path.exists(os.environ[TARGET_SCRIPT_ENV_KEY])):
            return os.environ[TARGET_SCRIPT_ENV_KEY]

        # Now try name matching.
        # Go up in stack, check frames with source files in current folder
        # and try to match their names with potential script names
        for frame in inspect.stack():
            path = frame[1]
            # if the file is in current folder
            if (os.path.exists(path)
                    and os.path.samefile(os.path.dirname(path), os.getcwd())
                    and path.endswith(".py")):
                tester_name = os.path.basename(path)[:-3]

                # see if there is a file without test prefix
                for prefix in _test_file_prefixes:
                    if tester_name.startswith(prefix):
                        target_name = tester_name[len(prefix):]
                        if os.path.exists(target_name + ".py"):
                            return os.path.join(os.getcwd(), target_name + ".py")

                # see if there is a file without test suffix
                for suffix in _test_file_suffixes:
                    if tester_name.endswith(suffix):
                        target_name = tester_name[:-len(suffix)]
                        if os.path.exists(target_name + ".py"):
                            return os.path.join(os.getcwd(), target_name + ".py")

        raise AssertionError(CANT_FIND_TARGET)


def _looks_like_test_file(path):
    name = os.path.basename(path)
    if not name.endswith(".py"):
        return False

    name = name[:-3]

    for prefix in _test_file_prefixes:
        if name.startswith(prefix):
            return True

    for suffix in _test_file_suffixes:
        if name.endswith(suffix):
            return True

    return False


def get_function(name, scope=None):
    """Tries to locate given function ..."""
    assert isinstance(name, str)

    if scope is None:
        if "." in name:
            scope, name = name.rsplit(".", maxsplit=1)
        else:
            scope = find_target()

    if isinstance(scope, str):
        scope = import_module(scope)

    if isinstance(scope, types.ModuleType):
        scope = vars(scope)

    assert isinstance(scope, dict)
    assert name in scope, CANT_FIND_FUNCTION.format(name=name)
    return scope[name]


def import_module(name, package=None):
    """Works like :func:`importlib.import_module` but raises :exc:ForbiddenInputError
    instead of blocking when module tries to read from stdin.
    """
    try:
        with ForbiddenInput():
            return importlib.import_module(name, package)
    except ForbiddenInputError:
        raise ImportError(CANT_IMPORT_MODULE_WITH_INPUT)


def check_function(function, arguments, expected_result, keyword_arguments={},
                   inputs=[],
                   allow_remaining_inputs=False,
                   expected_stdout=None,
                   expected_stderr="",
                   may_modify_globals=True,
                   may_modify_arguments=True):
    if isinstance(function, str):
        function = get_function(function)

    # prepare the call
    if not may_modify_globals:
        original_globals = deepcopy(vars(inspect.getmodule(function)))

    if not may_modify_arguments:
        # TODO: repr
        original_arguments = deepcopy(arguments)
        original_keyword_arguments = deepcopy(keyword_arguments)

    # make the call
    with IOCapturing(inputs) as iocap:
        try:
            actual_result = function(*arguments, **keyword_arguments)
        except OutOfInputsError:
            if inputs in ["", [], ()]:
                raise ForbiddenInputError(FUNCTION_READ_STDIN)
            else:
                raise

    stdout = iocap.get_stdout()
    stderr = iocap.get_stderr()

    # check result
    if actual_result != expected_result:

        # Maybe student used print instead of return?
        if (actual_result == None
                and str(expected_result).trim() == stdout.trim()):
            raise AssertionError(FUNCTION_PRINTED_INSTEAD_OF_RETURNING)

        if type(actual_result) != type(expected_result):
            raise AssertionError(FUNCTION_RETURNED_WRONG_TYPE.format(
                expected=type(expected_result).__name__,
                actual=type(actual_result).__name__))

        raise AssertionError(FUNCTION_RETURNED_WRONG_VALUE.format(
            expected=expected_result,
            actual=actual_result))

    if expected_stdout == "" and stdout != "":
        raise AssertionError(FUNCTION_WROTE_STDOUT)

    if expected_stderr == "" and stderr != "":
        raise AssertionError(FUNCTION_WROTE_STDERR)

    check_std_streams(iocap, expected_stdout, expected_stderr, allow_remaining_inputs)

    # check other conditions
    if (not may_modify_arguments
            and (original_arguments != arguments
                 or original_keyword_arguments != keyword_arguments)):
        raise AssertionError(FUNCTION_MODIFIED_ARGUMENTS)

    if not may_modify_globals and original_globals != vars(inspect.getmodule(function)):
        raise AssertionError(FUNCTION_MODIFIED_GLOBALS)


def check_pure_function(function, arguments, expected_result, keyword_arguments={}):
    """Shortcut for :func:`check_function` disallowing IO
    and modifications to globals and arguments. Note that it doesn't actually
    guarantee that function is indeed pure, it only checks for common reasons
    why it can't be pure."""
    return check_function(function, arguments, expected_result, keyword_arguments,
                          inputs=[], expected_output="",
                          may_modify_globals=False,
                          may_modify_arguments=False)


def check_script(script_path=None, inputs=[], arguments=[],
                 provided_files={},
                 text_file_encoding="UTF-8",
                 text_file_universal_newlines=True,
                 allow_remaining_inputs=False,
                 expected_stdout=None,
                 expected_last_stdout=None,
                 expected_stderr="",
                 expected_files={}
                 ):
    try:
        globs, iocap = run_script(script_path, inputs, arguments,
                                  provided_files, text_file_encoding)
        check_std_streams(iocap, expected_stdout, expected_stderr, allow_remaining_inputs)

        for path in expected_files:
            check_file_contains(path, expected_files[path],
                                text_file_encoding, text_file_universal_newlines)

        return globs

    finally:
        # remove temp files
        for path in provided_files.keys().union(expected_files.keys()):
            if os.path.exists(path):
                os.remove(path)


def check_file_contains(path, expected_content, text_file_encoding="UTF-8",
                        universal_newlines=True):
    assert os.path.exists(path), CANT_FIND_FILE.format(path=path)

    if isinstance(expected_content, str):
        expected_content = expected_content
        with open(path, encoding=text_file_encoding) as fp:
            actual_content = fp.read()

        if universal_newlines:
            expected_content = expected_content.replace("\r\n", "\n")
            actual_content = actual_content.replace("\r\n", "\n")

        if actual_content != expected_content:
            raise AssertionError("TODO: file content doesn't match")

    elif isinstance(expected_content, bytes):
        with open(path, mode="b") as fp:
            actual_content = fp.read()

        if actual_content != expected_content:
            raise AssertionError("TODO: file content doesn't match")

    elif expected_content is None:
        pass
    else:
        raise TypeError("Unsupported expected_content")


def check_std_streams(iocap, expected_stdout, expected_stderr, allow_remaining_inputs=False):
    "TODO:"


def check_code(obj,
               disallow_keywords=[],
               disallow_names=[]):
    pass


def run_script(script_path=None, inputs=[], arguments=[],
               provided_files={}, text_file_encoding="UTF-8"):
    """
    Returns pair with global variables and :class:IOCapturing object
    """
    if script_path is None:
        script_path = find_target()

    # preparations
    for path in provided_files:
        create_file(path, provided_files[path], text_file_encoding)

    original_argv = sys.argv
    sys.argv = [script_path] + arguments

    with IOCapturing(inputs) as iocap:
        try:
            # Execute code located at the specified filesystem location
            # Returns the resulting top level namespace dictionary
            globs = runpy.run_path(script_path, run_name="__main__")
        finally:
            sys.argv = original_argv

    return globs, iocap


def generate_function_test(function, arguments, expected_result):
    fun = get_function(function)

    title_text = fun.__name__ + "(" + ", ".join(map(repr, arguments)) + ")"

    @title(title_text)
    def test():
        check_function(fun, arguments, expected_result)

    # Assuming the caller of generate_function_test is in test module.
    # Attach test to this module.
    caller_frame = inspect.currentframe().f_back
    test_module = inspect.getmodule(caller_frame)
    setattr(test_module, title_text, test)


class Plan:
    """Context manager which adds plan description to exception messages"""

    def __init__(self, description):
        self._description = description

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            # TODO: show part of stacktrace
            # TODO: detect errors in the test
            if exc_type is AssertionError:
                msg = str(exc_value)
            else:
                msg = str(exc_type.__name__) + ": " + str(exc_value)

            raise AssertionError(self._description + "\n\n" + msg)


class ForbiddenIO():
    """Context manager which causes access to given standard streams
    to raise :exc:`ForbiddenOutputError` or :exc:`ForbiddenInputError`


    Example ::

        with ForbiddenIO({"stdin", "stderr"}):
            print("Blah 2")                         # OK, stdout is not forbidden
            something = input("input something: ")  # raises ForbiddenInputError

    """

    def __init__(self, forbidden_streams={"stdin", "stdout", "stderr"}):
        self._forbidden_streams = forbidden_streams
        self._original_streams = {}

    def __enter__(self):
        for name in self._forbidden_streams:
            self._original_streams[name] = getattr(sys, name)
            setattr(sys, name, _ForbiddenStream())

    def __exit__(self, exc_type, exc_value, traceback):
        for name in self._forbidden_streams:
            setattr(sys, name, self._original_streams[name])


class ForbiddenInput(ForbiddenIO):
    def __init__(self):
        ForbiddenIO.__init__(self, {"stdin"})


class IOCapturing:
    def __init__(self, inputs=[]):
        self._inputs = inputs

    def __enter__(self):
        self._inputs_iterator = self._create_inputs_iterator(self._inputs)
        self._stream_events = []
        self._original_streams = {}
        self._remaining_inputs = []

        # Remember original streams
        for stream_name in {"stdout", "stderr", "stdin"}:
            self._original_streams[stream_name] = getattr(sys, stream_name)

        # Install fake streams
        sys.stdout = _OutputListenerStream(self._record_stdout_data)
        sys.stderr = _OutputListenerStream(self._record_stderr_data)
        sys.stdin = _InputListenerStream(self._record_stdin_data, self._inputs_iterator)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._remaining_inputs = sys.stdin.fetch_remaining_inputs()

        # restore original streams
        for stream_name in {"stdout", "stderr", "stdin"}:
            setattr(sys, stream_name, self._original_streams[stream_name])

    def get_remaining_inputs(self):
        return self._remaining_inputs

    def get_io(self):
        return self._get_stream_data({"stdout", "stderr", "stdin"}, False)

    def get_stdout(self):
        return self._get_stream_data({"stdout"}, False)

    def get_last_stdout(self):
        return self._get_stream_data({"stdout"}, True)

    def get_stderr(self):
        return self._get_stream_data({"stderr"}, False)

    def debug(self, *args, sep=' ', end='\n', stream_name="stdout", flush=False):
        """Meant for printing debug information from input generator."""
        print(*args, sep=sep, end=end,
              file=self._original_streams[stream_name], flush=flush)

    def _record_stdout_data(self, data):
        self._stream_events.append(("stdout", data))

    def _record_stderr_data(self, data):
        self._stream_events.append(("stderr", data))

    def _record_stdin_data(self, data):
        self._stream_events.append(("stdin", data))

    def _get_stream_data(self, stream_names, only_since_last_input):
        result = ""
        i = len(self._stream_events) - 1
        while i >= 0:
            event_stream_name, event_data = self._stream_events[i]

            if only_since_last_input and event_stream_name == "stdin":
                break

            if event_stream_name in stream_names:
                result = event_data + result

            i -= 1

        return result

    def _create_inputs_iterator(self, inputs):
        if isinstance(inputs, str):
            return iter(inputs.splitlines())

        try:
            return iter(inputs)
        except TypeError:
            try:
                result = inputs(self)
                assert isinstance(result, types.GeneratorType)
                return result

            except BaseException as e:
                e.message = "Could not create iterator for inputs.\n" + e.message
                raise


class _ListenerStream:
    def __init__(self, data_listener):
        self._data_listener = data_listener


class _OutputListenerStream(_ListenerStream):
    def __init__(self, data_listener):
        _ListenerStream.__init__(self, data_listener)

    def write(self, data):
        self._data_listener(data)

    def writelines(self, lines):
        self._data_listener(lines)


class _InputListenerStream(_ListenerStream):

    def __init__(self, data_listener, inputs_iterator):
        _ListenerStream.__init__(self, data_listener)
        self._inputs_iterator = inputs_iterator

    def fetch_remaining_inputs(self):
        result = []
        while True:
            try:
                result.append(next(self._inputs_iterator))
            except StopIteration:
                break

        return result

    def read(self, limit=-1):
        raise NotImplementedError("Method 'read' on stdin is not supported in testable programs")

    def readline(self, limit=-1):
        if limit > 0:
            raise NotImplementedError("Positive limit/size for stdin.readline is not supported in testable programs")

        try:
            data = str(next(self._inputs_iterator)) + "\n"
            self._data_listener(data)
            return data
        except StopIteration:
            raise OutOfInputsError(OUT_OF_INPUTS)

    def readlines(self, limit=-1):
        raise NotImplementedError("Method 'readlines' on stdin is not supported in testable programs")


class ForbiddenStreamError(AssertionError):
    pass


class OutOfInputsError(RuntimeError):
    pass


class ForbiddenInputError(ForbiddenStreamError):
    pass


class ForbiddenOutputError(ForbiddenStreamError):
    pass


class _ForbiddenStream:
    def write(self, data):
        raise ForbiddenOutputError(OUTPUT_NOT_ALLOWED)

    def writelines(self, lines):
        raise ForbiddenOutputError(OUTPUT_NOT_ALLOWED)

    def read(self, limit=-1):
        raise ForbiddenInputError(INPUT_NOT_ALLOWED)

    def readline(self, limit=-1):
        raise ForbiddenInputError(INPUT_NOT_ALLOWED)

    def readlines(self, limit=-1):
        raise ForbiddenInputError(INPUT_NOT_ALLOWED)


def load_locale(locale):
    """Example: if argument is "et", then all global variables defined
    in edutest_locale_et.py are copied to edutest, overriding existing values"""
    try:
        this_module = sys.modules[__name__]
        that_module = importlib.import_module("edutest_locale_" + locale)

        for name in vars(that_module):
            if not name.startswith("_"):
                setattr(this_module, name, getattr(that_module, name))
    except:
        pass


def title(title_text):
    """Creates decorator for setting test title. Works only for Nose.
    Usage example:

    @title("First test with x=3 and y=4")
    def test1():
        assert fun(x=3, y=4) == 7

    """

    def decorator(f):
        f.__name__ = title_text
        f.__test__ = True
        return f

    return decorator


def get_source(path):
    import tokenize
    # encoding-safe open
    with tokenize.open(path) as sourceFile:
        contents = sourceFile.read()
    return contents


def is_function(value):
    try:
        return hasattr(value, '__call__')
    except:
        return False


def get_error_message(exception=None):
    if exception:
        type_ = type(exception)
        msg = str(exception)
    else:
        type_, msg, _ = sys.exc_info()

    return "{}: {}".format(type_.__name__, msg)


def get_traceback(exception):
    type_, value, tb = type(exception), exception, exception.__traceback__
    return "".join(traceback.format_exception(type_, value, tb))


def quote_text_block(text):
    return ("\n>" + text.replace("\n", "\n>")
            + ("" if text.endswith("\n") else "\n"))


def extract_numbers(s, allow_decimal_comma=False):
    result = []

    if allow_decimal_comma:
        rexp = """((?:\+|\-)?\d+(?:(?:\.|,)\d+)?)"""
    else:
        rexp = """((?:\+|\-)?\d+(?:\.\d+)?)"""

    for item in re.findall(rexp, s):
        try:
            result.append(int(item))
        except:
            try:
                result.append(float(item.replace(",", ".")))
            except:
                pass
    return result


def contains_number(num_list_or_str, x, allowed_error=0, allow_decimal_comma=False):
    if isinstance(num_list_or_str, str):
        nums = extract_numbers(num_list_or_str, allow_decimal_comma)
    else:
        nums = num_list_or_str

    for num in nums:
        if abs(num - x) <= allowed_error:
            return True

    return False


def ast_contains_name(node, name):
    if isinstance(node, ast.Name) and node.id == name:
        return True

    for child in ast.iter_child_nodes(node):
        if ast_contains_name(child, name):
            return True

    return False


def ast_contains(node, node_type):
    if isinstance(node, node_type):
        return True

    for child in ast.iter_child_nodes(node):
        if ast_contains(child, node_type):
            return True

    return False


def create_file(path, content, text_file_encoding=None):
    if text_file_encoding is None:
        text_file_encoding = locale.getpreferredencoding(False)

    if isinstance(content, str):
        content = content.encode(text_file_encoding)

    with open(path, mode="wb") as fp:
        fp.write(content)


def delete_if_exists(path):
    if os.path.exists(path):
        os.remove(path)


def check_file_exists(name):
    assert os.path.isfile(name), "Ei leia faili nimega %s" % name  # TODO: translate


def _hide_from_nose():
    # hide all functions from nose (there are several with "test" in name)
    for name in vars(sys.modules[__name__]).copy():
        if "test" in name.lower():
            try:
                setattr(getattr(sys.modules[__name__], name), "__test__", False)
            except:
                pass


def _load_preferred_locale():
    def get_language_code_from_locale_string(s):
        return s.split(".")[0].split("_")[0]

    if "VPL_LANG" in os.environ:  # locale set by VPL server (http://vpl.dis.ulpgc.es/)
        load_locale(get_language_code_from_locale_string(os.environ["VPL_LANG"]))
    else:
        try:
            load_locale(get_language_code_from_locale_string(locale.getdefaultlocale()[0]))
        except Exception as e:
            print(e)
            return


def description(desc):
    """Can be used for changing test function name"""

    def decorator(f):
        f.__name__ = desc
        return f

    return decorator


_load_preferred_locale()
_hide_from_nose()



from edutest import *
from inspect import signature
import ast
import os
from copy import deepcopy
import datetime


# Lõpetab automaatkontrollimise (st genereerib erindi)
def lõpeta(teade, väljund=""):
    if väljund != "":
        if teade[-1] != '\n':
            teade += '\n'
        teade += "\nProgrammi väljund oli:" + quote_text_block(väljund)
    raise AssertionError(teade)


# Abifunktsioon tekstide "arv+ühik" väljastamiseks
def arvsõne(n, ainsus, mitmus):
    s = str(n) + " "
    if n == 1:
        s += ainsus
    else:
        s += mitmus
    return s


# Abifunktsioon järjendi liikmete väljastamiseks
def jadasõne(a):
    s = ""
    for i in range(len(a)):
        if i > 0:
            s += " "
        s += str(a[i])
        if i < len(a) - 2:
            s += ","
        elif i == len(a) - 2:
            s += " ja"
    return s


# Kontrollib, kas sisend on arv
def on_arv(s):
    try:
        float(str(s))
        return True
    except ValueError:
        return False
    except TypeError:
        return False


# Kontrollib, kas sisend on täisarv
def on_täisarv(s):
    try:
        int(str(s))
        return True
    except ValueError:
        return False
    except TypeError:
        return False


# Kontrollib, kas antud nimega programm on olemas
def progolemasolu(programminimi):
    assert os.path.isfile(programminimi), "Ei leia faili %s." % programminimi
    assert os.path.getsize(programminimi) > 0, "Failil %s puudub sisu." % programminimi


# Programmi sisendparameetrite arvu kontrollimine ja teadete väljastamine.
# Veatekst on täiendav info kasutajale vea korral.
def progsisendiarvukontroll(tegelik_sisendite_arv, oodatav_sisendite_arv, veatekst=""):
    if tegelik_sisendite_arv != oodatav_sisendite_arv:
        if tegelik_sisendite_arv == 0 and oodatav_sisendite_arv > 0:
            teade = "Programm ei küsinud sisendeid.\n"
            teade += "Programm peaks küsima sisendit kasutajalt."
            lõpeta(teade)
        elif tegelik_sisendite_arv > 0 and oodatav_sisendite_arv == 0:
            teade = "Programm küsis sisendit.\n"
            teade += "Programm ei tohiks kasutajalt ühtegi sisendit küsida."
            lõpeta(teade)
        elif float('Inf') > tegelik_sisendite_arv > 0 and oodatav_sisendite_arv > 0:
            teade = "Programm küsis "
            teade += ("vähem" if tegelik_sisendite_arv < oodatav_sisendite_arv else "rohkem")
            teade += " sisendeid, kui oleks pidanud.\n"
            teade += "Programm küsis " + arvsõne(tegelik_sisendite_arv, "sisendit", "sisendit") \
                     + ", aga oleks pidanud küsima " + arvsõne(oodatav_sisendite_arv, "sisendit", "sisendit") + ".\n"
            teade += veatekst
            lõpeta(teade)
        elif tegelik_sisendite_arv == float('Inf') and oodatav_sisendite_arv > 0:
            teade = "Programm küsis rohkem sisendeid, kui oleks pidanud.\n"
            teade += "Programm peaks küsima ainult " + arvsõne(oodatav_sisendite_arv, "sisendit", "sisendit") + ".\n"
            teade += veatekst
            lõpeta(teade)


# Täidab programmi etteantud sisenditega, kontrollib ka sisendite arvu.
def täidaprog(programminimi, sisendid, oodatav_sisendite_arv, veatekst=""):
    try:
        globs, cap = run_script(programminimi, sisendid)
    except OutOfInputsError:
        tegelik_sisendite_arv = float('Inf')
    else:
        tegelik_sisendite_arv = len(sisendid) - len(cap.get_remaining_inputs())
    progsisendiarvukontroll(tegelik_sisendite_arv, oodatav_sisendite_arv, veatekst)
    return globs, cap


# Kontrollib, kas programmis on olemas antud nime ja parameetrite arvuga funktsioon.
def funksignatuur(globs, funktsiooninimi, oodatav_parameetrite_arv, veatekst=""):
    if funktsiooninimi not in globs:
        lõpeta("Ei leia programmist funktsiooni '" + str(funktsiooninimi) + "'.")

    tegelik_parameetrite_arv = len(signature(globs[funktsiooninimi]).parameters)
    if tegelik_parameetrite_arv != oodatav_parameetrite_arv:
        teade = "Funktsioonil '" + str(funktsiooninimi) + "' on " + \
                arvsõne(tegelik_parameetrite_arv, "parameeter", "parameetrit") + ", " \
                                                                                 "aga peaks olema " + \
                arvsõne(oodatav_parameetrite_arv, "parameeter", "parameetrit") + "."
        # Parameetrite kirjeldus
        if veatekst != "":
            teade += "\n" + veatekst
        lõpeta(teade)


# Funktsiooni sisendparameetrite arvu kontrollimine ja teadete väljastamine.
def funksisendiarvukontroll(funktsiooninimi, tegelik_sisendite_arv, oodatav_sisendite_arv, veatekst=""):
    if tegelik_sisendite_arv != oodatav_sisendite_arv:
        if tegelik_sisendite_arv == 0 and oodatav_sisendite_arv > 0:
            teade = "Funktsioon '" + str(funktsiooninimi) + "' ei küsinud sisendeid.\n"
            teade += "Funktsioon peaks küsima sisendit kasutajalt."
            lõpeta(teade)
        elif tegelik_sisendite_arv > 0 and oodatav_sisendite_arv == 0:
            teade = "Funktsioon '" + str(funktsiooninimi) + "' küsis "
            teade += ("sisendit" if tegelik_sisendite_arv == 1 else "sisendeid") + ".\n"
            teade += "Funktsioon ei tohiks kasutajalt ühtegi sisendit küsida, andmed tuleks talle ette anda argumentidena."
            lõpeta(teade)
        elif float('Inf') > tegelik_sisendite_arv > 0 and oodatav_sisendite_arv > 0:
            teade = "Funktsioon '" + str(funktsiooninimi) + "' küsis "
            teade += ("vähem" if tegelik_sisendite_arv < oodatav_sisendite_arv else "rohkem")
            teade += " sisendeid, kui oleks pidanud.\n"
            teade += "Funktsioon küsis " + arvsõne(tegelik_sisendite_arv, "sisendit", "sisendit") \
                     + ", aga oleks pidanud küsima " + arvsõne(oodatav_sisendite_arv, "sisendit", "sisendit") + ".\n"
            teade += veatekst
            lõpeta(teade)
        elif tegelik_sisendite_arv == float('Inf') and oodatav_sisendite_arv > 0:
            teade = "Funktsioon '" + str(funktsiooninimi) + "' küsis "
            teade += "rohkem sisendeid, kui oleks pidanud.\n"
            teade += "Funktsioon peaks küsima ainult " + arvsõne(oodatav_sisendite_arv, "sisendit", "sisendit") + ".\n"
            teade += veatekst
            lõpeta(teade)


# Täidab funktsiooni etteantud argumentide ja sisenditega, kontrollib ka sisendite arvu.
def täidafunk(funktsioon, argumendid, sisendid, oodatav_sisendite_arv, veatekst=""):
    try:
        with IOCapturing(sisendid) as cap:
            tegelik_tulemus = funktsioon(*argumendid)
    except OutOfInputsError:
        tegelik_sisendite_arv = float('Inf')
    else:
        tegelik_sisendite_arv = len(sisendid) - len(cap.get_remaining_inputs())
    funksisendiarvukontroll(funktsioon.__name__, tegelik_sisendite_arv, oodatav_sisendite_arv, veatekst)
    return tegelik_tulemus, cap


# Kontrollib, kas programm pöördub funktsiooni poole.
def progpöördub(programminimi, funktsiooninimi):
    with open(programminimi, encoding='utf-8') as f:
        programmisisu = f.read()
        süntaksipuu = ast.parse(programmisisu)
        if type(funktsiooninimi) == str:
            kontrollitavad = [funktsiooninimi]
        elif type(funktsiooninimi) == list:
            kontrollitavad = funktsiooninimi
        for tipp in ast.walk(süntaksipuu):
            if isinstance(tipp, ast.Call):
                if 'func' in tipp._fields:
                    if 'id' in tipp.func._fields:
                        for i in range(len(kontrollitavad)):
                            if tipp.func.id == kontrollitavad[i]:
                                return True
                    if 'attr' in tipp.func._fields:
                        for i in range(len(kontrollitavad)):
                            if tipp.func.attr == kontrollitavad[i]:
                                return True
    return False


# Kontrollib, kas funktsioon pöördub funktsiooni poole.
def funkpöördub(programminimi, funktsiooninimi, kontrollitavnimi):
    with open(programminimi, encoding='utf-8') as f:
        programmisisu = f.read()
        süntaksipuu = ast.parse(programmisisu)
        if type(kontrollitavnimi) == str:
            kontrollitavad = [kontrollitavnimi]
        elif type(kontrollitavnimi) == list:
            kontrollitavad = kontrollitavnimi
        for tipp in ast.walk(süntaksipuu):
            if isinstance(tipp, ast.FunctionDef) and tipp.name == funktsiooninimi:
                for alamtipp in ast.walk(tipp):
                    if isinstance(alamtipp, ast.Call):
                        if 'func' in alamtipp._fields:
                            if 'id' in alamtipp.func._fields:
                                for i in range(len(kontrollitavad)):
                                    if alamtipp.func.id == kontrollitavad[i]:
                                        return True
                            if 'attr' in alamtipp.func._fields:
                                for i in range(len(kontrollitavad)):
                                    if alamtipp.func.attr == kontrollitavad[i]:
                                        return True
    return False


# Annab veateate, kui programm ei kasuta funktsiooni.
def progfunkkasutamine(programminimi, funktsiooninimi):
    if not progpöördub(programminimi, funktsiooninimi):
        teade = "Programm ei kasuta funktsiooni '" + funktsiooninimi + "'.\n"
        teade += "Väljund tuleks arvutada programmis defineeritud funktsiooni '" + funktsiooninimi + "' abil."
        lõpeta(teade)


# Kontrollib, kas programm kasutab moodulit
def kasutabmoodulit(programminimi, moodulinimi):
    with open(programminimi, encoding='utf-8') as f:
        programmisisu = f.read()
    süntaksipuu = ast.parse(programmisisu)
    if type(moodulinimi) == str:
        kontrollitavad = [moodulinimi]
    elif type(moodulinimi) == list:
        kontrollitavad = moodulinimi
    for tipp in ast.walk(süntaksipuu):
        if isinstance(tipp, ast.Import):
            if 'names' in tipp._fields:
                for nimi in tipp.names:
                    if 'name' in nimi._fields:
                        for i in range(len(kontrollitavad)):
                            if nimi.name == kontrollitavad[i]:
                                return True
        if isinstance(tipp, ast.ImportFrom):
            if 'module' in tipp._fields:
                for i in range(len(kontrollitavad)):
                    if tipp.module == kontrollitavad[i]:
                        return True
    return False


# Annab veateate, kui programm ei kasuta moodulit.
def moodulikasutamine(programminimi, moodulinimi):
    if not kasutabmoodulit(programminimi, moodulinimi):
        teade = "Programm ei kasuta moodulit '" + moodulinimi + "'.\n"
        lõpeta(teade)


# Kontrollib, kas funktsioon püüab tagastamise asemel väärtust väljastada.
def tühiprintteade(programminimi, funktsiooninimi):
    teade = "Funktsioon '" + str(funktsiooninimi) + "' ei tagastanud väärtust."
    if funkpöördub(programminimi, funktsiooninimi, "print"):
        teade += "\nFunktsioon peaks väärtuse tagastama, mitte väljastama."
    lõpeta(teade)


# Kontrollib, kas funktsioon tagastas õiget tüüpi väärtuse.
def funktüüp(funktsiooninimi, tegelik_tulemus, õige_tüüp):
    if type(tegelik_tulemus) != õige_tüüp:
        teade = "Funktsioon '" + str(funktsiooninimi) + "' tagastas väärtuse, mis ei ole "
        if õige_tüüp == str:
            teade += "sõne"
        elif õige_tüüp == list:
            teade += "järjend"
        elif õige_tüüp == int:
            teade += "täisarv"
        elif õige_tüüp == bool:
            teade += "tõeväärtus"
        elif õige_tüüp == dict:
            teade += "sõnastik"
        elif õige_tüüp == tuple:
            teade += "ennik"
        else:
            teade += "õiget tüüpi"
        teade += ".\n"
        teade += "Tagastatud väärtus oli "
        if type(tegelik_tulemus) == str: teade += "'"
        teade += str(tegelik_tulemus)
        if type(tegelik_tulemus) == str: teade += "'"
        teade += "."
        lõpeta(teade)




from tiivad.tests import TestResult


def calculate_test_status(checks):
    for check_result in checks:
        if check_result.check_status == TestResult.FAIL.name:
            return TestResult.FAIL.name
    return TestResult.PASS.name


def format_checks(checks):
    result = []
    for check_result in checks:
        check = {}
        check["title"] = check_result.before_message
        check["status"] = check_result.check_status
        check["feedback"] = check_result.passed_message if check_result.check_status == TestResult.PASS.name \
            else check_result.failed_message
        result.append(check)

    return result


class Results(object):
    __instance = None
    tests = []
    pre_evaluate_error = None
    result_type = "OK_V3"
    points = None
    error = None

    def __new__(cls, val):
        if Results.__instance is None:
            Results.__instance = object.__new__(cls)
        if val is not None:
            Results.__instance.tests.append(val)
        return Results.__instance

    def clear(cls):
        cls.__instance = None

    def calculate_points(self):
        points = 0
        for test_result in self.tests:
            points += test_result.get("points", 0) \
                if test_result.get("test_status", TestResult.FAIL.name) == TestResult.PASS.name else 0
        return points

    def format_tests(self):
        tests = []
        for test_result in self.tests:
            result = {}
            result["title"] = test_result.get("name", "")
            result["status"] = calculate_test_status(test_result.get("checks", []))
            result["exception_message"] = test_result.get("exception_message", None)
            result["user_inputs"] = test_result.get("standard_input_data", [])
            result["created_files"] = [{"name": x[0], "content": x[1]} for x in test_result.get("input_files", [])]
            result["actual_output"] = test_result.get("actual_output", None)
            result["checks"] = format_checks(test_result.get("checks", []))
            tests.append(result)
        return tests

    def format_result(self):

        if self.pre_evaluate_error is not None:
            self.tests = []

        # TODO (kspar): tiivad should return (print) just the error message, not the wrapping message
        if self.error is not None:
            return {
                "message_type": "ERROR_V3",
                "message": {"error": self.error}
            }

        self.points = self.calculate_points()
        test_results = self.format_tests()

        result = {}
        result["result_type"] = self.result_type
        result["points"] = self.points
        result["pre_evaluate_error"] = self.pre_evaluate_error
        result["tests"] = test_results

        return result


if __name__ == '__main__':
    print(Results({"a", "b"}).tests)
    print(Results(None).tests)
    print(Results({"a", "b"}).tests)
    print(Results(None).tests)
    print(Results({"a", "b"}).tests)
    print(Results(None).tests)
    print(Results({"a", "b"}).tests)
    print(Results(None).tests)



'''
Kasutajalt küsitakse liini pikkust (täisarvuna meetrites)
ja kõrvutiasetsevate postide maksimaalkaugust (täisarvuna
meetrites). Ekraanile väljastatakse, mitu posti on liini
ehitamiseks minimaalselt vaja.
'''

# Kasutame moodulist math ülespoole ümardamise käsku ceil
from math import ceil

# Küsime kasutaja käest andmeid ja teisendame need täisarvudeks
liini_pikkus = int(input("Sisesta elektriliini pikkus täisarvuna meetrites: "))
postide_kaugus = int(input("Sisesta kõrvutasuvate postide maksimaalkaugus: "))

# Valem postide arvu leidmiseks
postide_arv = ceil(liini_pikkus / postide_kaugus) + 1

# Väljastame vastuse ekraanile
print("Liini ehitamiseks läheb vaja minimaalselt", postide_arv, "posti.")

u = "v"

def test(test: int):
    print("a")
    a = test - 3
    # nottest()
    b = 2
    print("c")
    s = []
    v = input("user input 1")
    w = input("user input 1")
    x = input("user input 1")
    y = input("user input 1")
    for i in s:
        pass
    return b

print("u")
test(1)


from tiivad import *

validate_files(['''kodu1.py'''])
execute_test(file_name='''kodu1.py''', standard_output_checks=[
    StandardOutputChecks(string_check_type='''ALL_OF_THESE''', nothing_else=None, expected_output=['''turtle'''],
                         before_message='''Kontrollib, kas programm impordib mooduli.''',
                         passed_message='''Programm importis mooduli.''',
                         failed_message='''Programm ei importinud moodulit.''')],
             type='''program_imports_module_test''', points=10.0, id=None, name='''Programm importib mooduli turtle.''',
             inputs=None, passed_next=None, failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', standard_output_checks=[
    StandardOutputChecks(string_check_type='''ANY_OF_THESE''', nothing_else=None,
                         expected_output=['''down''', '''up''', '''forward''', '''left''', '''right''', '''fd''',
                                          '''rt''', '''lt''', '''bk''', '''goto'''],
                         before_message='''Kontrollib, kas programm kutsub välja turtle funktsiooni.''',
                         passed_message='''Programm kutsus välja vähemalt ühe turtle funktsiooni.''',
                         failed_message='''Programm ei kutsnud välja ühtegi turtle funktsiooni.''')],
             type='''program_calls_function_test''', points=1.0, id=None,
             name='''Programm kutsub välja turtle funktsiooni.''', inputs=None, passed_next=None, failed_next=None,
             visible_to_user=None)
execute_test(file_name='''kodu1.py''', contains_check=(True, False),
             before_message='''Kontrollib, kas programmis esineb tsükkel.''',
             passed_message='''Programmis esines tsükkel.''', failed_message='''Programmis ei esine ühtegi tsüklit.''',
             type='''program_contains_loop_test''', points=1.0, id=None, name='''Programmis esineb tsükkel.''',
             inputs=None, passed_next=None, failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', contains_check=(False, True),
             before_message='''Kontrollib, kas programmis esineb try/except plokk.''',
             passed_message='''Programmis ei esine try/except plokki.''',
             failed_message='''Programmis esines try/except plokk.''', type='''program_contains_try_except_test''',
             points=1.0, id=None, name='''Programmis ei tohi olla try/except plokki.''', inputs=None, passed_next=None,
             failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', contains_check=(True, False),
             before_message='''Kontrollib, kas programm kutsub välja 'print' käsu.''',
             passed_message='''Programm kutsus välja 'print' käsu.''',
             failed_message='''Programm ei kutsunud välja 'print' käsku.''', type='''program_calls_print_test''',
             points=1.0, id=None, name='''Programm peab välja kutsuma 'print' käsu.''', inputs=None, passed_next=None,
             failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', standard_output_checks=[
    StandardOutputChecks(string_check_type='''ANY_OF_THESE''', nothing_else=None,
                         expected_output=['''down''', '''up''', '''seosta_lapsed_ja_vanemad'''],
                         before_message='''Kontrollib, kas programm defineerib vähemalt ühe antud funktsioonidest.''',
                         passed_message='''Programm defineeris vähemalt ühe antud funktsioonidest.''',
                         failed_message='''Programm ei defineerinud välja ühtegi antud funktsioonidest.''')],
             type='''program_defines_function_test''', points=1.0, id=None,
             name='''Programm defineerib vähemalt ühe antud funktsioonidest.''', inputs=None, passed_next=None,
             failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', standard_output_checks=[
    StandardOutputChecks(string_check_type='''ANY_OF_THESE''', nothing_else=None,
                         expected_output=['''seosta_lapsed_ja_vanemad''', '''yes''', '''no'''],
                         consider_elements_order=None,
                         before_message='''Kontrollib, kas programm sisaldab võtmesõna.''',
                         passed_message='''Programm sisaldab võtmesõna.''',
                         failed_message='''Programm ei sisalda võtmesõna.''')],
             type='''program_contains_keyword_test''', points=1.0, id=None, name='''Programm sisaldab võtmesõna.''',
             inputs=None, passed_next=None, failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', standard_input_data=['''Kasutajasisend'''],
             input_files=[('''sisendfail.txt''', '''Sisendfaili sisu.''')], standard_output_checks=[
        StandardOutputChecks(string_check_type='''ANY_OF_THESE''', nothing_else=None,
                             expected_output=['''down''', '''up'''], consider_elements_order=False,
                             before_message='''Kontrollib, kas väljund sisaldab võtmesõna.''',
                             passed_message='''Programmi väljund sisaldab antud võtmesõna.''',
                             failed_message='''Programmi väljund ei sisalda antud võtmesõna.''')], output_file_checks=[
        OutputFileChecks(file_name='''valjundfail.txt''', string_check_type='''ALL_OF_THESE''', nothing_else=None,
                         expected_output=['''Väljundfaili eeldatav sisu'''], consider_elements_order=False,
                         before_message='''Kontrollib, kas väljundfail sisaldab võtmesõna.''',
                         passed_message='''Programmi väljundfail sisaldab antud võtmesõna.''',
                         failed_message='''Programmi väljundfail ei sisalda antud võtmesõna.''')],
             exception_check=(False, True), before_message=None, passed_message=None, failed_message=None,
             type='''program_execution_test''', points=1.0, id=None, name='''Programmi käivituse test.''', inputs=None,
             passed_next=None, failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', function_name='''seosta_lapsed_ja_vanemad''', contains_check=(True, False),
             before_message='''Kontrollib, kas funktsioonis esineb tsükkel.''',
             passed_message='''Funktsioonis esines tsükkel.''',
             failed_message='''Funktsioonis ei esine ühtegi tsüklit.''', type='''function_contains_loop_test''',
             points=1.0, id=None, name='''Funktsioonis esineb tsükkel.''', inputs=None, passed_next=None,
             failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', function_name='''seosta_lapsed_ja_vanemad''', contains_check=(True, False),
             before_message='''Kontrollib, kas funktsioonis esineb return.''',
             passed_message='''Funktsioonis esines return.''',
             failed_message='''Funktsioonis ei esine ühtegi return'i.''', type='''function_contains_return_test''',
             points=1.0, id=None, name='''Funktsioonis esineb return.''', inputs=None, passed_next=None,
             failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', function_name='''seosta_lapsed_ja_vanemad''', standard_output_checks=[
    StandardOutputChecks(string_check_type='''ANY_OF_THESE''', nothing_else=None,
                         expected_output=['''down''', '''up''', '''forward''', '''left''', '''right''', '''fd''',
                                          '''rt''', '''lt''', '''bk''', '''goto'''],
                         before_message='''Kontrollib, kas funktsioon kutsub välja turtle funktsiooni.''',
                         passed_message='''Funktsioon kutsus välja vähemalt ühe turtle funktsiooni.''',
                         failed_message='''Funktsioon ei kutsnud välja ühtegi turtle funktsiooni.''')],
             type='''function_calls_function_test''', points=1.0, id=None,
             name='''Funktsioon kutsub välja turtle funktsiooni.''', inputs=None, passed_next=None, failed_next=None,
             visible_to_user=None)
execute_test(file_name='''kodu1.py''', function_name='''seosta_lapsed_ja_vanemad''', contains_check=(True, False),
             before_message='''Kontrollib, kas funktsioon kutsub välja print käsu.''',
             passed_message='''Funktsioon kutsub välja print käsu.''',
             failed_message='''Funktsioon ei kutsu välja print käsku.''', type='''function_calls_print_test''',
             points=1.0, id=None, name='''Funktsioon kutsub välja print käsu.''', inputs=None, passed_next=None,
             failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', function_name='''seosta_lapsed_ja_vanemad''', contains_check=(True, False),
             before_message='''Kontrollib, kas funktsioon on rekursiivne.''',
             passed_message='''Funktsioonis on rekursiivne.''', failed_message='''Funktsioonis ei ole rekursiivne.''',
             type='''function_is_recursive_test''', points=1.0, id=None, name='''Funktsioonis on rekursiivne.''',
             inputs=None, passed_next=None, failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', function_name='''seosta_lapsed_ja_vanemad''', standard_output_checks=[
    StandardOutputChecks(string_check_type='''ANY_OF_THESE''', nothing_else=None,
                         expected_output=['''down''', '''up''', '''seosta_lapsed_ja_vanemad2'''],
                         before_message='''Kontrollib, kas funktsioon defineerib vähemalt ühe antud funktsioonidest.''',
                         passed_message='''Funktsioon defineeris vähemalt ühe antud funktsioonidest.''',
                         failed_message='''Funktsioon ei defineerinud ühtegi antud funktsioonidest.''')],
             type='''function_defines_function_test''', points=1.0, id=None,
             name='''Funktsioon defineerib vähemalt ühe antud funktsioonidest.''', inputs=None, passed_next=None,
             failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', function_name='''seosta_lapsed_ja_vanemad''', standard_output_checks=[
    StandardOutputChecks(string_check_type='''ALL_OF_THESE''', nothing_else=None, expected_output=['''turtle'''],
                         before_message='''Kontrollib, kas funktsioon impordib mooduli.''',
                         passed_message='''Funktsioon importis mooduli.''',
                         failed_message='''Funktsioon ei importinud moodulit.''')],
             type='''function_imports_module_test''', points=1.0, id=None,
             name='''Funktsioon importib mooduli turtle.''', inputs=None, passed_next=None, failed_next=None,
             visible_to_user=None)
execute_test(file_name='''kodu1.py''', function_name='''seosta_lapsed_ja_vanemad''', contains_check=(False, True),
             before_message='''Kontrollib, kas funktsioonis esineb try/except plokk.''',
             passed_message='''Funktsioonis ei esine try/except plokki.''',
             failed_message='''Funktsioonis esines try/except plokk.''', type='''function_contains_try_except_test''',
             points=1.0, id=None, name='''Funktsioonis ei tohi olla try/except plokki.''', inputs=None,
             passed_next=None, failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', function_name='''seosta_lapsed_ja_vanemad''', contains_check=(True, False),
             before_message='''Kontrollib, kas funktsioon kasutab ainult lokaalseid muutujaid.''',
             passed_message='''Funktsioon kasutab ainult lokaalseid muutujaid.''',
             failed_message='''Funktsioon ei kasuta ainult lokaalseid muutujaid.''',
             type='''function_uses_only_local_vars_test''', points=1.0, id=None,
             name='''Funktsioon kasutab ainult lokaalseid muutujaid.''', inputs=None, passed_next=None,
             failed_next=None, visible_to_user=None)
execute_test(file_name='''kodu1.py''', function_name='''seosta_lapsed_ja_vanemad''',
             arguments=['''lapsed.txt''', '''nimed.txt'''], standard_input_data=['''Kasutajasisend'''],
             input_files=[('''sisendfail.txt''', '''Sisendfaili sisu.''')], return_value='''{}''',
             standard_output_checks=[StandardOutputChecks(string_check_type='''ANY_OF_THESE''', nothing_else=None,
                                                          expected_output=['''down''', '''up'''],
                                                          consider_elements_order=False,
                                                          before_message='''Kontrollib, kas funktsiooni väljund sisaldab võtmesõna.''',
                                                          passed_message='''Funktsiooni väljund sisaldab antud võtmesõna.''',
                                                          failed_message='''Funktsiooni väljund ei sisalda antud võtmesõna.''')],
             output_file_checks=[OutputFileChecks(file_name='''valjundfail.txt''', string_check_type='''ALL_OF_THESE''',
                                                  nothing_else=None, expected_output=['''Väljundfaili eeldatav sisu'''],
                                                  consider_elements_order=False,
                                                  before_message='''Kontrollib, kas funktsiooni väljundfail sisaldab võtmesõna.''',
                                                  passed_message='''Funktsiooni väljundfail sisaldab antud võtmesõna.''',
                                                  failed_message='''Funktsiooni väljundfail ei sisalda antud võtmesõna.''')],
             type='''function_execution_test''', points=1.0, id=None, name='''Funktsiooni käivituse test.''',
             inputs=None, passed_next=None, failed_next=None, visible_to_user=None)
print(json.dumps(Results(None).format_result(), cls=ComplexEncoder, ensure_ascii=False))



class Check:
    before_message = ""
    passed_message = ""
    failed_message = ""
    check_status = ""
    error_message = ""


class StandardOutputChecks:
    def __init__(self, string_check_type, nothing_else, expected_output, before_message, passed_message, failed_message,
                 consider_elements_order=None):
        self.string_check_type = string_check_type
        self.nothing_else = nothing_else
        self.expected_output = expected_output
        self.before_message = before_message
        self.passed_message = passed_message
        self.failed_message = failed_message
        self.check_status = None
        self.error_message = None
        self.consider_elements_order = consider_elements_order

    def __str__(self):
        return {
            "string_check_type": self.string_check_type,
            "nothing_else": self.nothing_else,
            "expected_output": self.expected_output,
            "before_message": self.before_message,
            "passed_message": self.passed_message,
            "failed_message": self.failed_message,
            "check_status": self.check_status,
            "error_message": self.error_message,
            "consider_elements_order": self.consider_elements_order
        }

    def __repr__(self):
        return {
            "string_check_type": self.string_check_type,
            "nothing_else": self.nothing_else,
            "expected_output": self.expected_output,
            "before_message": self.before_message,
            "passed_message": self.passed_message,
            "failed_message": self.failed_message,
            "check_status": self.check_status,
            "error_message": self.error_message,
            "consider_elements_order": self.consider_elements_order
        }.__str__()

    def repr_json(self):
        return {
            "string_check_type": self.string_check_type,
            "nothing_else": self.nothing_else,
            "expected_output": self.expected_output,
            "before_message": self.before_message,
            "passed_message": self.passed_message,
            "failed_message": self.failed_message,
            "check_status": self.check_status,
            "error_message": self.error_message,
            "consider_elements_order": self.consider_elements_order
        }


class OutputFileChecks:
    def __init__(self, file_name, string_check_type, nothing_else, expected_output, before_message, passed_message, failed_message,
                 consider_elements_order=None):
        self.file_name = file_name
        self.string_check_type = string_check_type
        self.nothing_else = nothing_else
        self.expected_output = expected_output
        self.before_message = before_message
        self.passed_message = passed_message
        self.failed_message = failed_message
        self.check_status = None
        self.error_message = None
        self.consider_elements_order = consider_elements_order

    def __str__(self):
        return {
            "file_name": self.file_name,
            "string_check_type": self.string_check_type,
            "nothing_else": self.nothing_else,
            "expected_output": self.expected_output,
            "before_message": self.before_message,
            "passed_message": self.passed_message,
            "failed_message": self.failed_message,
            "check_status": self.check_status,
            "error_message": self.error_message,
            "consider_elements_order": self.consider_elements_order
        }

    def __repr__(self):
        return {
            "file_name": self.file_name,
            "string_check_type": self.string_check_type,
            "nothing_else": self.nothing_else,
            "expected_output": self.expected_output,
            "before_message": self.before_message,
            "passed_message": self.passed_message,
            "failed_message": self.failed_message,
            "check_status": self.check_status,
            "error_message": self.error_message,
            "consider_elements_order": self.consider_elements_order
        }.__str__()
